import contextlib as _contextlib
import logging as _logging
import os as _os
import warnings as _warnings
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

from transformers.tokenization_utils import AddedToken as _AddedToken
from transformers.tokenization_utils import PreTrainedTokenizer as _PreTrainedTokenizer

from tamm.tokenizers.sentencepiece import _add_user_defined_tokens
from tamm.typing import PathLike as _PathLike
from tamm.utils.uri import _is_uri, _URIHandler

_HFTokenType = _Union[str, _AddedToken]

_logger = _logging.getLogger(__name__)


class TammSentencePieceTokenizer(_PreTrainedTokenizer):
    """
    This is a HuggingFace tokenizer implementation that wraps an underlying
    :module:`sentencepiece` tokenizer.  Compared to the :class:`LlamaTokenizer`
    from :module:`transformers`, it is a thinner wrapper around the SentencePiece
    model, and it intentionally does not support ``split_special_tokens=False``
    (at least for now). We do this so that the tokenizer more precisely follows
    the behavior of the underlying SentencePiece tokenizer.

    Args:
        vocab_file (:obj:`str`): A filepath or URI for the SentencePiece
            vocab file.
        bos_token (:obj:`str` or :obj:`AddedToken`): The beginning-of-sequence
            token, typically ``"<s>"``.
        eos_token (:obj:`str` or :obj:`AddedToken`): The end-of-sequence
            token, typically ``"</s>"``.
        unk_token (:obj:`str` or :obj:`AddedToken`): The unknown token,
            typically ``"<unk>"``.
        pad_token (:obj:`str` or :obj:`AddedToken`): The padding token,
            typically ``"<pad>"``.
        additional_special_tokens (:obj:`list` of :obj:`str`): An optional
            list of additional special tokens.
        add_bos_token (:obj:`bool`, optional): A flag that controls prepending
            the BOS token id when encoding.  Defaults to ``False``.  When
            enabled, prepending only happens if ``add_special_tokens`` is also
            ``True``.
        chat_template (:obj:`str`): An optional Jinja template for
            :meth:`apply_chat_template`.
        chat_template_strip_first_token (:obj:`str`, optional): A flag that
            controls whether to drop the first token when calling
            :meth:`apply_chat_template`.  Defaults to ``False``.
        chat_template_add_special_tokens (:obj:`str`, optional): A flag that
            controls whether to add special tokens in :meth:`apply_chat_template`.
            Defaults to ``False``.
        text_replacements (:obj:`dict` of :obj:`str` pairs): A dictionary
            of text pairs to preprocess encoding and postprocess decoding.
            For example, ``text_replacements={"\n": "<n>"}`` replaces
            newlines with ``"<n>"`` prior to tokenization and replaces
            "<n>" with newlines after detokenization.
        **kwargs: Additional keyword arguments for
            :meth:`PreTrainedTokenizer.__init__`.
    """

    # Developer note: The point of this class is to provide a HF tokenizer that
    # can be configured to precisely match the SentencePiece behavior that we run
    # in production.  As of transformers==4.50, this is tricky due to some tedious
    # details, mostly involving HF's custom treatment of special tokens when
    # split_special_tokens=False (the HF default).  In particular, this option
    # results in different handling of the SentencePiece control tokens <s>, </s>,
    # <pad>, and <unk>, and it also requires many tricks to preserve the
    # SentencePiece dummy prefix behavior (see the implementation of
    # transformers.LlamaTokenizer).  We work around these challenges by dynamically
    # adding user-defined tokens to the SentencePiece vocab and forcing
    # split_special_tokens=True.  The only downside that we are aware of is that
    # some HF special token features (such as stripping neighboring whitespace) are
    # not supported.  This is not concerning, since SentencePiece does not support
    # this behavior either.

    vocab_files_names = {"vocab_file": "tokenizer.model"}
    model_input_names = ["input_ids", "attention_mask"]

    def __init__(
        self,
        vocab_file: _Union[str, _PathLike],
        bos_token: _Optional[_HFTokenType] = "<s>",
        eos_token: _Optional[_HFTokenType] = "</s>",
        unk_token: _Optional[_HFTokenType] = "<unk>",
        pad_token: _Optional[_HFTokenType] = "<pad>",
        additional_special_tokens: _Optional[_List[_HFTokenType]] = None,
        add_bos_token: bool = False,
        chat_template: _Optional[str] = None,
        chat_template_add_special_tokens: bool = False,
        chat_template_strip_first_token: bool = False,
        text_replacements: _Optional[_Dict[str, str]] = None,
        **kwargs,
    ):
        self.vocab_file = vocab_file
        self.text_replacements = text_replacements
        self.add_bos_token = add_bos_token

        self.chat_template_add_special_tokens = chat_template_add_special_tokens
        self.chat_template_strip_first_token = chat_template_strip_first_token
        self._is_apply_chat_template_context = False

        if additional_special_tokens is None:
            additional_special_tokens = []
        self._init_sp_model(vocab_file)
        self._register_sp_tokens(
            bos_token, eos_token, unk_token, pad_token, *additional_special_tokens
        )

        kwargs.setdefault("split_special_tokens", True)
        super().__init__(
            bos_token=bos_token,
            eos_token=eos_token,
            unk_token=unk_token,
            pad_token=pad_token,
            additional_special_tokens=additional_special_tokens,
            add_bos_token=add_bos_token,
            chat_template=chat_template,
            chat_template_add_special_tokens=chat_template_add_special_tokens,
            chat_template_strip_first_token=chat_template_strip_first_token,
            text_replacements=text_replacements,
            **kwargs,
        )

    def _init_sp_model(self, vocab_file: str = None) -> None:
        import sentencepiece  # pylint: disable=import-outside-toplevel

        self._sp_model = sentencepiece.SentencePieceProcessor()
        if vocab_file is None:
            return

        if _is_uri(vocab_file):
            vocab_file = _URIHandler(use_cache=True).map_to_local(vocab_file)
        self._sp_model.Load(str(vocab_file))

    def _register_sp_tokens(self, *new_tokens):
        current_vocab = self.get_vocab()
        new_tokens = [str(tok) for tok in new_tokens]
        new_tokens = [
            tok for tok in new_tokens if tok is not None and tok not in current_vocab
        ]
        if len(new_tokens) == 0:
            return
        new_tokens = list(dict.fromkeys(new_tokens))  # dedupe, preserving order
        _add_user_defined_tokens(self._sp_model, new_tokens)

    @property
    def vocab_size(self) -> int:
        return self._sp_model.get_piece_size()

    def get_vocab(self) -> _Dict[str, int]:
        return {
            self._sp_model.id_to_piece(token_id): token_id
            for token_id in range(self.vocab_size)
        }

    def add_tokens(
        self,
        new_tokens: _Union[_HFTokenType, _List[_HFTokenType]],
        special_tokens: bool = False,
        **kwargs,
    ) -> int:
        """
        Wraps :meth:`transformers.PreTrainedTokenizer.add_tokens` to also
        add the new tokens to the SentencePiece model's vocabulary.
        """
        if not isinstance(new_tokens, (list, tuple)):
            new_tokens = [new_tokens]
        self._register_sp_tokens(*new_tokens)
        return super().add_tokens(new_tokens, special_tokens=special_tokens, **kwargs)

    def prepare_for_tokenization(
        self, text: str, is_split_into_words: bool = False, **kwargs
    ) -> _Tuple[str, _Dict[str, _Any]]:
        """
        Performs string replacements prior to tokenization.  See
        :meth:`transformers.PreTrainedTokenizer.prepare_for_tokenization` for
        more info.
        """
        if self.text_replacements is not None:
            for old, new in self.text_replacements.items():
                text = text.replace(old, new)
        return text, kwargs

    @property
    def split_special_tokens(self) -> bool:
        """
        A flag that controls the :meth:`tokenize` behavior.  When ``False``,
        :meth:`tokenize` applies special treatment to special tokens.  This
        class forces the flag to ``True``.
        """
        return True

    @split_special_tokens.setter
    def split_special_tokens(self, value: bool) -> None:
        if not value:
            _warnings.warn(
                "split_special_tokens assigned to False in SentencePieceTokenizer, "
                "which is unsupported.  Forcing this value to True.  Please contact "
                "the tamm team if you require support for this feature."
            )

    def tokenize(self, text: str, **kwargs) -> _List[str]:
        """
        Wraps :meth:`transformers.tokenize` to force ``split_special_tokens=True``.

        Args:
            text (:obj:`str`): The text to tokenize.
            **kwargs: Keyword arguments for the wrapped
                :meth:`transformers.PreTrainedTokenizer.tokenize` method.
        """
        split_special_tokens = kwargs.pop("split_special_tokens", True)
        if not split_special_tokens:
            _warnings.warn(
                "SentencePieceTokenizer.tokenize() is ignoring False value for"
                "split_special_tokens, which is unsupported.  Please contact the"
                "tamm team if you require support for this feature."
            )
        result = super().tokenize(text, **kwargs)
        if (
            self._is_apply_chat_template_context
            and self.chat_template_strip_first_token
        ):
            result = result[1:]
        return result

    def __call__(self, *args, **kwargs):
        if self._is_apply_chat_template_context:
            kwargs["add_special_tokens"] = self.chat_template_add_special_tokens
        return super().__call__(*args, **kwargs)

    def build_inputs_with_special_tokens(
        self, token_ids_0: _List[int], token_ids_1: _Optional[_List[int]] = None
    ) -> _List[int]:
        prefix = []
        if self.add_bos_token:
            bos_id = self._convert_token_to_id(str(self.bos_token))
            prefix.append(bos_id)

        if prefix:
            token_ids_0 = prefix + token_ids_0

        if token_ids_1 is None:
            return token_ids_0

        return token_ids_0 + prefix + token_ids_1

    def get_special_tokens_mask(
        self,
        token_ids_0: _List[int],
        token_ids_1: _Optional[_List[int]] = None,
        already_has_special_tokens: bool = False,
    ) -> _List[int]:
        # Developer note (TJ): It's unclear to me how this is used within the HF ecosystem
        # (could not find an example easily).  This implementation follows LlamaTokenizer.

        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0,
                token_ids_1=token_ids_1,
                already_has_special_tokens=True,
            )
        sequences = [token_ids_0]
        if token_ids_1 is not None:
            sequences.append(token_ids_1)

        result = []
        for seq in sequences:
            if self.add_bos_token:
                result.append(1)
            result.extend([0] * len(seq))
        return result

    def _tokenize(self, text: str, **kwargs) -> _List[str]:
        return self._sp_model.encode_as_pieces(text)

    def _convert_token_to_id(self, token: str) -> int:
        return self._sp_model.piece_to_id(token)

    @_contextlib.contextmanager
    def _chat_template_context(self):
        original = self._is_apply_chat_template_context
        self._is_apply_chat_template_context = True
        yield
        self._is_apply_chat_template_context = original

    def apply_chat_template(self, *args, **kwargs):
        """
        Wraps :meth:`transformers.PreTrainedTokenizer.apply_chat_template` to
        pass additional ``tokenizer_kwargs`` to :meth:`.tokenize`.
        """
        with self._chat_template_context():
            return super().apply_chat_template(*args, **kwargs)

    def decode(
        self,
        token_ids: _Union[int, _List[int]],
        skip_special_tokens: bool = False,
        clean_up_tokenization_spaces: bool = None,
        **kwargs,
    ) -> str:
        """
        Wraps :meth:`transformers.PreTrainedTokenizer.decode` to
        apply string replacements after detokenization.
        """
        result = super().decode(
            token_ids,
            skip_special_tokens=skip_special_tokens,
            clean_up_tokenization_spaces=clean_up_tokenization_spaces,
            **kwargs,
        )
        if self.text_replacements is not None:
            for new, old in reversed(self.text_replacements.items()):
                result = result.replace(old, new)
        return result

    def convert_tokens_to_string(self, tokens: _List[str]) -> str:
        """
        Decodes a list of pieces to a string using the SentencePiece model.
        See :meth:`transformers.convert_tokens_to_string` for more info.
        """
        return self._sp_model.decode(tokens)

    def _convert_id_to_token(self, index: int) -> str:
        return self._sp_model.id_to_piece(index)

    def save_vocabulary(
        self, save_directory: str, filename_prefix: _Optional[str] = None
    ) -> _Tuple[str]:
        filename = self.vocab_files_names["vocab_file"]
        if filename_prefix is not None:
            filename = f"{filename_prefix}-{filename}"
        filepath = _os.path.join(save_directory, filename)
        file_contents = self._sp_model.serialized_model_proto()
        with open(filepath, "wb") as fp:
            fp.write(file_contents)
        return (filepath,)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_sp_model"] = None
        state["_sp_model_proto"] = self._sp_model.serialized_model_proto()
        return state

    def __setstate__(self, d):
        model_proto = d.pop("_sp_model_proto")
        self.__dict__.update(d)
        self._init_sp_model()
        self._sp_model.LoadFromSerializedProto(model_proto)



"""
layers.norm
-----------

This module provides norm layers and a factory function for creating norm layers.

.. autoclass:: tamm.layers.BatchNorm
    :show-inheritance:

.. autoclass:: tamm.layers.LayerNorm
    :show-inheritance:

.. autoclass:: tamm.layers.RMSNorm
    :show-inheritance:

.. autofunction:: tamm.layers.norm.create_norm_builder
"""

import abc as _abc
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn
from torch.nn.modules.batchnorm import _BatchNorm as _PyTorchBatchNorm

from tamm.layers import functional as _tamm_F
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class _BaseNorm(_nn.Module, _abc.ABC):
    """
    A Base class for norm layers.

    Args:
        features_shape (:obj:`tuple` of :obj:`int`): The shape of the feature (channel)
            dimension(s).
        bias (:obj:`bool`): A flag for including a bias parameter.  Defaults to
            ``False``.
        dim (:obj:`int` or :obj:`tuple` of :obj:`int`): The feature dimension index
            (or indices) to scale using the ``weight`` parameter.  Defaults to ``-1``,
            corresponding to channels-last inputs.  Use ``1`` for channels-first inputs.
        device: The device of the parameters.
        dtype: The dtype of the parameters.
    """

    _IMPL_CHANNELS_DIM = -1
    """The index of the features dimension for :meth:`_forward_impl`."""

    def __init__(
        self,
        features_shape: _Optional[_Iterable[int]] = None,
        *,
        dim: _Union[int, _Iterable[int]] = -1,
        eps: float = 1e-5,
        bias: bool = False,
        device=None,
        dtype=None,
    ):
        _nn.Module.__init__(self)

        if features_shape is not None:
            features_shape = tuple(features_shape)
        self._features_shape = features_shape

        if not isinstance(dim, int):
            dim = tuple(dim)
        self.dim = dim

        self.eps = eps

        if self._features_shape is not None:
            self.weight = _nn.Parameter(
                _torch.empty(self._features_shape, device=device, dtype=dtype)
            )
        else:
            self.weight = None

        if bias:
            self.bias = _nn.Parameter(
                _torch.empty(self._features_shape, device=device, dtype=dtype)
            )
        else:
            self.bias = None

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.weight is not None:
            _nn.init.ones_(self.weight)
        if self.bias is not None:
            _nn.init.zeros_(self.bias)

    def _permute(
        self, x: _torch.Tensor
    ) -> _Tuple[_torch.Tensor, _Optional[_List[int]]]:
        """
        Permutes the dimensions which will be normalized
        on to dimensions start from self._IMPL_CHANNELS_DIM.
        Returns a tuple with the permuted tensors and a list of indices
        which indicate how the permutation was performed. If the `_IMPL_CHANNELS_DIM`
        is a single dimension, returns None for the second return value.
        """
        if self.dim == self._IMPL_CHANNELS_DIM:
            return x, None
        if isinstance(self.dim, int):
            return x.transpose(self.dim, self._IMPL_CHANNELS_DIM), None

        # Permute multiple dimensions to the last dimensions
        norm_dim_list = sorted([d if d >= 0 else d + len(x.shape) for d in self.dim])
        remained_dim_list = [
            idx for idx in range(len(x.shape)) if idx not in norm_dim_list
        ]
        perm = remained_dim_list + norm_dim_list
        return _torch.permute(x, perm), perm

    def _unpermute(
        self, x: _torch.Tensor, perm: _Optional[_Iterable[int]] = None
    ) -> _torch.Tensor:
        if self.dim == self._IMPL_CHANNELS_DIM:
            return x
        if isinstance(self.dim, int):
            return x.transpose(self.dim, self._IMPL_CHANNELS_DIM)
        return _tamm_F.inverse_permute(x, perm)

    @_abc.abstractmethod
    def _forward_impl(self, x: _torch.Tensor) -> _torch.Tensor:
        """Normalize x assuming the features dimension of x is _IMPL_CHANNELS_DIM"""

    # pylint: disable=redefined-builtin
    def forward(self, input: _torch.Tensor) -> _torch.Tensor:
        x, perm = self._permute(input)
        x = self._forward_impl(x)
        return self._unpermute(x, perm)


class LayerNorm(_BaseNorm, _LayerMixin):
    """LayerNorm with optional bias param."""

    def __init__(
        self,
        normalized_shape: _Iterable[int],
        *,
        dim: _Union[int, _Iterable[int]] = -1,
        eps: float = 1e-5,
        bias: bool = False,
        cast_dtype: _Optional[_torch.dtype] = _torch.float32,
        device=None,
        dtype=None,
    ):
        super().__init__(
            features_shape=normalized_shape,
            dim=dim,
            eps=eps,
            bias=bias,
            device=device,
            dtype=dtype,
        )
        self.cast_dtype = cast_dtype

    def _forward_impl(self, x: _torch.Tensor) -> _torch.Tensor:
        return _tamm_F.layer_norm(
            x,
            normalized_shape=self._features_shape,
            weight=self.weight,
            bias=self.bias,
            eps=self.eps,
            cast_dtype=self.cast_dtype,
        )

    def extra_repr(self) -> str:
        shape = tuple(self._features_shape)
        return (
            f"normalized_shape={shape}, dim={self.dim}, eps={self.eps}, "
            f"bias={self.bias is not None}, cast_dtype={self.cast_dtype}"
        )


class RMSNorm(_BaseNorm, _LayerMixin):
    """RMSNorm with no bias param."""

    def __init__(
        self,
        normalized_shape: _Iterable[int],
        *,
        dim: _Union[int, _Iterable[int]] = -1,
        eps: float = 1e-5,
        bias: bool = False,
        cast_dtype: _Optional[_torch.dtype] = _torch.float32,
        device=None,
        dtype=None,
    ):
        super().__init__(
            features_shape=normalized_shape,
            dim=dim,
            eps=eps,
            bias=bias,
            device=device,
            dtype=dtype,
        )
        self.cast_dtype = cast_dtype

    def _forward_impl(self, x: _torch.Tensor) -> _torch.Tensor:
        x = _tamm_F.rms_norm(
            x,
            self._features_shape,
            weight=self.weight,
            bias=self.bias,
            eps=self.eps,
            cast_dtype=self.cast_dtype,
        )
        return x

    def extra_repr(self) -> str:
        shape = tuple(self._features_shape)
        return (
            f"normalized_shape={shape}, dim={self.dim}, eps={self.eps}, "
            f"bias={self.bias is not None}, cast_dtype={self.cast_dtype}"
        )


class VectorizedRMSNorm(RMSNorm):
    def __init__(
        self,
        normalized_shape: _Iterable[int],
        vec_dim: int,
        vec_indx: int,
        *,
        dim: _Union[int, _Iterable[int]] = -1,
        eps: float = 1e-5,
        bias: bool = False,
        cast_dtype: _Optional[_torch.dtype] = _torch.float32,
        device=None,
        dtype=None,
    ):
        # TODO: actually vectorize this function
        new_shape = (vec_dim,) + tuple(normalized_shape)
        super(RMSNorm, self).__init__(
            features_shape=new_shape,
            dim=dim,
            eps=eps,
            bias=bias,
            device=device,
            dtype=dtype,
        )

        self.vec_dim = vec_dim
        self.vec_indx = vec_indx
        self.cast_dtype = cast_dtype

    # pylint: disable=redefined-builtin
    def forward(self, input: _torch.Tensor) -> _torch.Tensor:
        to_shape = list(
            ((1,) * (len(input.shape) - len(self._features_shape) + 1))
            + self._features_shape[1:]
        )
        to_shape[self.vec_indx] = self.vec_dim
        return _tamm_F.batched_rms_norms(
            input,
            self._features_shape[1:],
            weight=self.weight.reshape(to_shape),
            bias=None if self.bias is None else self.bias.reshape(to_shape),
            cast_dtype=self.cast_dtype,
        )


class BatchNorm(_BaseNorm, _LayerMixin, _PyTorchBatchNorm):
    """BatchNorm."""

    # Inherit from _PyTorchBatchNorm to extend functionality.
    #
    # Specifically, torch uses isinstance(module, _PyTorchBatchNorm),
    # in FSDP MixedPrecision to exclude some layers from low-precision
    # computation.
    #
    # We want this to return True for our BatchNorm layer too,
    # to exclude it from low precision computation.

    def __init__(
        self,
        features_shape: _Iterable[int],
        *,
        dim: _Optional[int] = -1,
        momentum: float = 0.1,
        eps: float = 1e-5,
        bias: bool = False,
        cast_dtype: _Optional[_torch.dtype] = _torch.float32,
        device=None,
        dtype=None,
    ):
        super().__init__(
            features_shape=features_shape,
            dim=dim,
            eps=eps,
            bias=bias,
            device=device,
            dtype=dtype,
        )
        self.momentum = momentum
        self.cast_dtype = cast_dtype
        self.register_buffer("running_mean", _torch.zeros_like(self.weight))
        self.register_buffer("running_var", _torch.ones_like(self.weight))
        self.register_buffer("num_batches_tracked", _torch.tensor(0, device=device))

    def _forward_impl(self, x: _torch.Tensor) -> _torch.Tensor:
        dim = (self.dim,) if not isinstance(self.dim, tuple) else self.dim

        # Flatten features then move features to dimension 1
        x, shape = x.flatten(start_dim=-len(dim)), x.shape[-len(dim) :]
        running_mean = self.running_mean.flatten()
        running_var = self.running_var.flatten()
        weight = self.weight.flatten()
        bias = self.bias.flatten() if self.bias is not None else None
        x = x.transpose(-1, 1)

        # Batch norm
        if self.training:
            self.num_batches_tracked.add_(1)
        normed_x = _tamm_F.batch_norm(
            x,
            running_mean=running_mean,
            running_var=running_var,
            weight=weight,
            bias=bias,
            training=self.training,
            momentum=self.momentum,
            eps=self.eps,
            cast_dtype=self.cast_dtype,
        )

        # Move features back to last dimension and unflatten features
        normed_x = normed_x.transpose(-1, 1)
        return normed_x.unflatten(self._IMPL_CHANNELS_DIM, shape)

    def extra_repr(self) -> str:
        shape = tuple(self._features_shape)
        return (
            f"features_shape={shape}, momentum={self.momentum}, dim={self.dim}, "
            f"eps={self.eps}, bias={self.bias is not None}, "
            f"cast_dtype={self.cast_dtype}"
        )

    @property
    def track_running_stats(self):
        return True

    @property
    def affine(self):
        return True


class L2Norm(_BaseNorm, _LayerMixin):
    def __init__(
        self,
        *,
        dim: _Optional[int] = -1,
        eps: float = 1e-5,
        cast_dtype: _Optional[_torch.dtype] = _torch.float32,
    ):
        super().__init__(dim=dim, eps=eps)
        self.cast_dtype = cast_dtype

    def _forward_impl(self, x: _torch.Tensor):
        return _tamm_F.l2_norm(x, eps=self.eps, cast_dtype=self.cast_dtype)

    def extra_repr(self) -> str:
        return f"eps={self.eps}, cast_dtype={self.cast_dtype}, dim={self.dim}"


def create_norm_builder(
    features_shape: _Tuple[int],
    spec: _Union[str, _OptionalModuleOrBuilder] = "layer_norm",
    *,
    bias: bool = False,
    dim: int = -1,
    eps: float = 1e-5,
    device=None,
    dtype=None,
):
    """
    Creates and returns a builder for a norm layer.

    Args:
        features_shape (:obj:`tuple` of :obj:`int`): The shape of the feature (channel)
            dimension(s).
        spec:
            Typically a :obj:`str` from the choices ``"layer_norm"``, ``"rms_norm"``,
            or ``"batch_norm"``.  This can also be ``None`` or a module builder,
            in which case the function returns spec directly.
        dim (:obj:`int` or :obj:`tuple` of :obj:`int`): The feature dimension index
            (or indices) to scale using the ``weight`` parameter.  Defaults to ``-1``,
            corresponding to channels-last inputs.  Use ``1`` for channels-first inputs.
        eps (:obj:`float`): The norm's epsilon parameter.  Defaults to ``1e-5``.
        bias (:obj:`bool`): A flag for including a bias parameter.  Defaults to
            ``False``.
        device: The device for parameters.
        dtype: The dtype for parameters.
    """
    if not isinstance(spec, str):
        return spec

    kwargs = {"dim": dim, "bias": bias, "eps": eps, "device": device, "dtype": dtype}
    if spec == "batch_norm":
        return BatchNorm.Builder(features_shape, **kwargs)
    if spec == "layer_norm":
        return LayerNorm.Builder(features_shape, **kwargs)
    if spec == "rms_norm":
        return RMSNorm.Builder(features_shape, **kwargs)
    raise ValueError(f"spec {spec} not recognized")



"""
layers.embedding
----------------

This module provides embedding layers.

.. autoclass:: tamm.layers.ConcatEmbedding
    :show-inheritance:
    :members: forward, create_builder_with_embedding

.. autoclass:: tamm.layers.ConstantEmbedding
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.ConvEmbedding
    :show-inheritance:
    :members:

.. autoclass:: tamm.layers.Embedding
    :show-inheritance:
    :members:

.. autoclass:: tamm.layers.UnionEmbedding
    :show-inheritance:
    :members: forward
"""

import logging as _logging
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn

from tamm.layers import basic as _basic
from tamm.layers import common as _layers_common
from tamm.layers import convolution as _convolution
from tamm.layers import functional as _tamm_F
from tamm.layers import init as _init
from tamm.layers import linear as _linear
from tamm.layers import sequential as _sequential
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder

_logger = _logging.getLogger(__name__)


class Embedding(_nn.Embedding, _LayerMixin):
    """
    Extension of :class:`torch.nn.Embedding` with several extra features:

    * Support for multidimensional embeddings by specifying a :obj:`tuple` for
      ``embedding_dim``.
    * Variance-scaled parameter initialization.
    * A :meth:`resize` method for easily changing the vocab size.

    The args are the same as :class:`torch.nn.Embedding` except
    ``embedding_dim`` may also be a tuple.  In this case, the layer outputs
    a tensor with shape ``(input_length, *embedding_dim)``.
    """

    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: _Union[int, _Tuple[int, ...]],
        padding_idx: _Optional[int] = None,
        max_norm: _Optional[float] = None,
        norm_type: float = 2.0,
        scale_grad_by_freq: bool = False,
        sparse: bool = False,
        device=None,
        dtype=None,
    ):
        try:
            embedding_dim = int(embedding_dim)
            self.embedding_shape = (embedding_dim,)
        except (ValueError, TypeError):
            self.embedding_shape = tuple(embedding_dim)
            embedding_dim = _torch.Size(self.embedding_shape).numel()
        super().__init__(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
            padding_idx=padding_idx,
            max_norm=max_norm,
            norm_type=norm_type,
            scale_grad_by_freq=scale_grad_by_freq,
            sparse=sparse,
            device=device,
            dtype=dtype,
        )

    def forward(self, *args, **kwargs):  # pylint: disable=useless-parent-delegation
        x = super().forward(*args, **kwargs)
        if len(self.embedding_shape) > 1:
            x = x.reshape(-1, *self.embedding_shape)
        return x

    def reset_parameters(self) -> None:
        _init.shape_normalized_normal_(self.weight, dim=1)

    def resize(
        self,
        new_num_embeddings: _Optional[int] = None,
        pad_to_multiple_of: _Optional[int] = None,
    ):
        """
        Resizes the token embedding layer.

        Args:
            new_num_embeddings (:obj:`_Optional[int]`): The number of tokens for
                resizing the input token embeddings layer.
            pad_to_multiple_of (:obj:`_Optional[int]`): The value for padding to
                a multiple of when resizing the input token embeddings layer.

        Returns:
            The resized embedding layer.
        """

        if pad_to_multiple_of is not None:
            if not isinstance(pad_to_multiple_of, int):
                raise ValueError(
                    f"Asking to pad the embedding matrix to a multiple of "
                    f"`{pad_to_multiple_of}`, which is not an integer. "
                    f"Please ensure you provide an integer value."
                )
            if new_num_embeddings is None:
                new_num_embeddings = self.weight.shape[0]
            new_num_embeddings = (
                (new_num_embeddings + pad_to_multiple_of - 1) // pad_to_multiple_of
            ) * pad_to_multiple_of

        if new_num_embeddings is None:
            return self

        if self.num_embeddings == new_num_embeddings:
            return self

        new_embeddings = Embedding(
            new_num_embeddings,
            self.embedding_dim,
            device=self.weight.device,
            dtype=self.weight.dtype,
        )

        # numbers of tokens to copy
        n = min(self.num_embeddings, new_num_embeddings)

        # copy token embeddings from the previous weights
        new_embeddings.weight.data[:n, :] = self.weight.data[:n, :]

        old_embeddings_requires_grad = self.weight.requires_grad
        new_embeddings.requires_grad_(old_embeddings_requires_grad)
        self.weight = new_embeddings.weight
        self.num_embeddings = new_embeddings.num_embeddings

        return self

    def __repr__(self):
        rep = super().__repr__()
        if len(self.embedding_shape) > 1:
            pieces = rep.split(",")
            end = ")" if len(pieces) == 2 else ""
            pieces[1] = f" {self.embedding_shape}{end}"
            rep = ",".join(pieces)
        return rep


class ConcatEmbedding(_nn.Module, _LayerMixin):
    """
    A layer that concatenates embeddings to the start and/or end of its inputs.

    Args:
        dim (:obj:`int`): The embedding dimension.
        num_start (:obj:`int`): The number of embeddings to prepend to the layer's
            inputs.  Defaults to 0.
        num_end (:obj:`int`): The number of embeddings to append to the layer's inputs.
            Defaults to 0.
        device: The device for parameters.
        dtype: The type for parameters.
    """

    def __init__(
        self, dim: int, *, num_start: int = 0, num_end: int = 0, device=None, dtype=None
    ):
        super().__init__()

        self.dim = dim
        self.num_start = num_start
        self.num_end = num_end

        if num_start > 0:
            self.weight_start = _torch.nn.Parameter(
                _torch.empty(num_start, dim, device=device, dtype=dtype)
            )
        else:
            self.weight_start = None

        if num_end > 0:
            self.weight_end = _torch.nn.Parameter(
                _torch.empty(num_end, dim, device=device, dtype=dtype)
            )
        else:
            self.weight_end = None

        self.reset_parameters()

    def reset_parameters(self):
        if self.weight_start is not None:
            _init.shape_normalized_normal_(self.weight_start, dim=1)
        if self.weight_end is not None:
            _init.shape_normalized_normal_(self.weight_end, dim=1)

    def extra_repr(self):
        return f"dim={self.dim}, num_start={self.num_start}, " f"num_end={self.num_end}"

    def forward(self, inputs: _torch.Tensor) -> _torch.Tensor:
        """
        Concatenates the layer's embeddings to the start and end of ``inputs`` along
        the sequence dimension.

        Args:
            inputs (:obj:`torch.Tensor`): Input embeddings with shape
                ``(batch size, seq len, dim)``.
        """
        embeddings = [inputs]
        if self.weight_start is not None:
            embeddings.insert(
                0, self._transform_to_match_inputs(self.weight_start, inputs=inputs)
            )
        if self.weight_end is not None:
            embeddings.append(
                self._transform_to_match_inputs(self.weight_end, inputs=inputs)
            )
        if len(embeddings) == 1:
            return embeddings[0]
        return _torch.cat(embeddings, dim=1)

    @staticmethod
    def _transform_to_match_inputs(embeddings, *, inputs):
        """Updates embeddings to match the dtype and batch size of inputs"""
        embeddings = embeddings.type_as(inputs)
        if inputs.size(0) == 1:
            return embeddings[None, ...]  # pylint: disable=unsubscriptable-object
        return embeddings.tile(inputs.size(0), 1, 1)

    @classmethod
    def create_builder_with_embedding(
        cls,
        embedding_builder: _layers_common.LayerBuilder,
        *,
        embedding_name: str = "embedding",
        **kwargs,
    ):
        """
        Converts a builder for an embedding layer into a builder for a sequence of
        two layers: the original embedding and a :obj:`.ConcatEmbedding`.

        Args:
            embedding_builder (:obj:`LayerBuilder`): A builder for an embedding layer.
            embedding_name (:obj:`str`, optional): A name for the embedding layer.
                Defaults to ``"embedding"``.
            kwargs: Keyword arguments for :class:`.ConcatEmbedding`.
        """
        builder = cls.Builder(**kwargs)
        return _sequential.Sequential.Builder(
            {embedding_name: embedding_builder, "concat_embedding": builder}
        )


class ConstantEmbedding(_nn.Module, _LayerMixin):
    """
    A layer that returns trainable embeddings with a fixed shape.

    Args:
        num_embeddings (:obj:`int`): The number of trainable embeddings.
        dim (:obj:`int`): The dimension of each embedding.
        device: The device for parameters.
        dtype: The dtype for parameters.
    """

    def __init__(
        self,
        num_embeddings: int,
        dim: int,
        *,
        device=None,
        dtype=None,
    ):
        super().__init__()

        self.num_embeddings = num_embeddings
        self.dim = dim

        self.weight = _torch.nn.Parameter(
            _torch.empty(num_embeddings, dim, device=device, dtype=dtype)
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        _init.shape_normalized_normal_(self.weight, dim=1)

    def extra_repr(self) -> str:
        return f"num_embeddings={self.num_embeddings}, dim={self.dim}"

    def forward(self, inputs: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            inputs (:obj:`torch.Tensor`): An input tensor where the first dimension
                is the batch size.

        Returns:
            A tensor with shape ``(batch_size, num_embeddings, dim)`` (the embeddings
            replicated along the batch dimension).
        """
        return _tamm_F.add_batch_dim(self.weight, batch_size=inputs.size(0))


class ConvEmbedding(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    A layer for extracting embeddings via a convolution.  This is a convolution layer
    followed by optional positional encoding and rearrange layers (typically a
    :class:`.ChannelsFirstToLast`).
    """

    def __init__(
        self,
        *,
        convolution: _ModuleOrBuilder,
        positional_encoding: _OptionalModuleOrBuilder = None,
        rearrange: _OptionalModuleOrBuilder = None,
        norm: _OptionalModuleOrBuilder = None,
    ):
        layers = {
            "convolution": convolution,
            "positional_encoding": positional_encoding,
            "rearrange": rearrange,
            "norm": norm,
        }
        super().__init__(layers)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        input_dim: int,
        output_dim: int,
        kernel_size: _Union[_Tuple[int, ...], int],
        stride: _Optional[_Union[_Tuple[int, ...], int]] = None,
        bias: bool = False,
        apply_positional_embedding: bool = False,
        positional_embedding_shape: _Tuple[int, ...] = None,
    ) -> _layers_common.LayerBuilder:
        """
        Creates and returns a builder for a :obj:`ConvEmbedding` that maps 2D images
        in NCHW format to embeddings.

        Args:
            input_dim (:obj:`int`): The number of input channels to the convolution.
            output_dim (:obj:`int`): The dimension of the output embeddings.
            kernel_size (:obj:`tuple` of :obj:`int`): The kernel size for the
                convolution.
            stride (:obj:`tuple` of :obj:`int`, optional): The stride for the
                convolution.  If ``None``, this defaults to the kernel size.
            bias (:obj:`bool`, optional): A flag for including a bias parameter in the
                convolution when ``True``.  Defaults to ``False``.
            apply_positional_embedding (:obj:`bool`, optional): A flag for including a
                :class:`SpatialPositionalEmbedding` layer.  Defaults to ``False``.
            positional_embedding_shape (:obj:`tuple` of :obj:`int`): The spatial shape
                for the positional embedding layer if
                ``apply_positional_embedding=True``.

        Returns:
            The newly created :obj:`LayerBuilder`.
        """
        if stride is None:
            stride = kernel_size
        conv = _convolution.Conv2d.Builder(
            in_channels=input_dim,
            out_channels=output_dim,
            kernel_size=kernel_size,
            stride=stride,
            bias=bias,
        )

        if apply_positional_embedding:
            # pylint: disable-next=import-outside-toplevel,cyclic-import
            from tamm.layers import positional_encoding as _positional_encoding

            pos_encoding = _positional_encoding.SpatialPositionalEmbedding.Builder(
                positional_embedding_shape, dim=output_dim
            )
        else:
            pos_encoding = None

        rearrange = _basic.ChannelsFirstToLast.Builder()

        return cls.Builder(
            convolution=conv, positional_encoding=pos_encoding, rearrange=rearrange
        )


class UnionEmbedding(_basic.Union, _LayerMixin):
    """
    A layer for composing multiple child embedding layers.

    Args:
        named_layers (:obj:`dict`): A dictionary that maps the name of each embedding
            layer to a builder or :obj:`nn.Module`.
        flatten_outputs (:obj:`bool`): A flag for flattening outputs of the layer.
            Defaults to ``False``.
    """

    def __init__(
        self,
        named_layers: _Dict[str, _OptionalModuleOrBuilder],
        *,
        flatten_outputs: bool = False,
    ):
        super().__init__(named_layers)
        self.flatten_outputs = flatten_outputs

    def extra_repr(self) -> str:
        return f"flatten_outputs={self.flatten_outputs}"

    def forward(self, inputs: _Any) -> _Union[_torch.Tensor, _Dict[str, _Any]]:
        """
        Args:
            inputs: Inputs to the child embedding layers.

        Returns:
            If ``flatten_outputs=False``, returns a :obj:`dict` that maps the name of
            each child layer to its output.  If ``flatten_outputs=True``, returns a
            single tensor with shape ``(batch_size, sequence_len, dim)``.  This flat
            tensor results from flattening the sequence dimensions of each child output
            and then concatenating the outputs along the sequence dim in the order
            of the children.
        """
        result = super().forward(inputs)
        if self.flatten_outputs:
            return _tamm_F.maybe_flatten_embeddings(result)
        return result


class LowRankFactorizedEmbedding(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    Low rank embeddings to reduce overhead associated with increased vocab.
    """

    def __init__(self, embedding: _ModuleOrBuilder, projection: _ModuleOrBuilder):
        named_layers = {"embedding": embedding, "projection": projection}
        super().__init__(named_layers)

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        num_embeddings: int,
        rank: int,
        output_dim: int,
        padding_idx: _Optional[int] = None,
    ):
        if rank > output_dim:
            raise ValueError(f"rank={rank} exceeds output_dim={output_dim}")
        embedding = Embedding.Builder(
            num_embeddings=num_embeddings, embedding_dim=rank, padding_idx=padding_idx
        )
        projection = _linear.Linear.Builder(rank, output_dim, bias=False)
        return cls.Builder(embedding=embedding, projection=projection)



from tamm.hf.transformers.models.sentencepiece.configuration_sentencepiece import (
    TammSentencePieceConfig,
)
from tamm.hf.transformers.models.sentencepiece.tokenization_sentencepiece import (
    TammSentencePieceTokenizer,
)



"""
tamm._cli
---------

This module implements a command-line interface for tamm.

To add new commands, we use the :func:`command` and `argument` decorators.  Here is a
toy example of how we could add a command that prints a string:

.. code-block:: python

    @command("examples", "print")
    @argument("--uppercase", action="store_true")
    @argument("string")
    def print_string(string, uppercase):
        if uppercase:
            string = string.upper()
        print(string)


As long as this code is imported during ``tamm._cli`` import, we can run it as
follows:

::

    > tamm examples print my_string
    my_string
    > tamm examples print "new string" --uppercase
    NEW STRING


For each cli command, we have an argparse subparser and a function to run when the
user invokes the command.  The subparser defines arguments for the command as well as
a ``func`` default argument with the command function as the value.  To run the cli, we
parse the args and then call ``func`` with the arguments.
"""
from tamm._cli import clean, from_tamm, ls, show, to_tamm, version
from tamm._cli.common import argument, command



"""
transformer.attention_mask
^^^^^^^^^^^^^^^^^^^^^^^^^^

This module implements attention mask layers.

.. autoclass:: AttentionMask
    :members: forward
"""

from typing import Any as _Any
from typing import Dict as _Dict
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm.layers import flash_attention as _flash
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import token_metadata as _token_metadata


class AttentionMask(_nn.Module, _LayerMixin):
    """
    A layer that maps segment IDs and positions into an attention mask for restricting
    attention between tokens.

    Args:
        is_causal (:obj:`bool`, optional):  A flag that specifies a causal mask pattern
            when ``True`` (i.e., a query token only attends to a key if the query
            position is at least as large as the key position).  If ``False``, the
            mask does not depend on the token positions.  Defaults to ``False``.
        flash_attention_mode (:obj:`str` or :obj:`.FlashAttentionMode`, optional): An
            enum for controlling the integration with FlashAttention.  Defaults to
            ``auto``.
    """

    def __init__(
        self,
        is_causal: bool = False,
        flash_attention_mode: _Union[str, _flash.FlashAttentionMode] = "auto",
    ):
        super().__init__()
        self.is_causal = is_causal

        if flash_attention_mode is None:
            flash_attention_mode = _flash.FlashAttentionMode.NONE
        self.flash_attention_mode = _helpers.get_enum_member_from_name(
            _flash.FlashAttentionMode, flash_attention_mode
        )
        _flash.is_flash_attention_available()

    def forward(
        self,
        token_metadata: _token_metadata.TokenMetadata,
        *,
        compute_dtype: _torch.dtype = _torch.float32,
    ) -> _Dict[str, _Any]:
        """
        Args:
            token_metadata (:obj:`.TokenMetadata`): The segment IDs, positions, and
                types of tokens for self attention.
            compute_dtype (:obj:`torch.dtype`, optional): The dtype for attention
                computation in the SDPA layer.  This option is only important if using
                FlashAttention.

        Returns:
            A :obj:`dict` with an ``"attention_mask"`` entry, which is a boolean tensor
            value of shape ``(batch size, seq_len, kv_seq_len)``.  A ``False`` value
            in the mask prevents corresponding tokens from attending to one another.
            The result also includes a ``"flash_attention_options"`` entry, which may
            have value ``None`` if FlashAttention is unavailable or disabled.
        """

        tm = token_metadata

        self._validate_positions(positions=tm.q_positions, kv_positions=tm.kv_positions)

        segment_mask = tm.q_segment_ids[..., None] == tm.kv_segment_ids[:, None, :]
        if self.is_causal:
            attention_mask = tm.q_positions[..., None] >= tm.kv_positions[:, None, :]
            attention_mask = _torch.logical_and(attention_mask, segment_mask)
        else:
            attention_mask = segment_mask

        # pylint: disable-next=assignment-from-none
        flash_options = _flash.compute_flash_attention_options(
            segment_ids=tm.q_segment_ids,
            kv_segment_ids=tm.kv_segment_ids,
            attention_mask=attention_mask,
            segment_mask=segment_mask,
            is_causal=self.is_causal,
            compute_dtype=compute_dtype,
            mode=self.flash_attention_mode,
        )

        return {
            "attention_mask": attention_mask,
            "flash_attention_options": flash_options,
        }

    def _validate_positions(self, *, positions, kv_positions):
        if not self.is_causal:
            return  # positions only used for causal masks
        if not _torch.is_tensor(positions):
            raise ValueError(
                "AttentionMask requires a positions tensor arg when is_causal=True"
            )
        if not _torch.is_tensor(kv_positions):
            raise ValueError(
                "AttentionMask requires a kv_positions tensor arg when is_causal=True"
            )

    def extra_repr(self) -> str:
        flash_mode = self.flash_attention_mode.name.lower()
        return f"is_causal={self.is_causal}, " f"flash_attention_mode='{flash_mode}'"



import logging as _logging

from tamm.runtime_configuration import rc as _rc


def get_logger(name: str = _rc.PROJECT_SLUG):
    return _logging.getLogger(name)


def set_logger_formatter(logger):
    try:
        import coloredlogs  # pylint: disable=import-outside-toplevel

        formatter = coloredlogs.ColoredFormatter(fmt=_rc.log_format)
    except Exception as e:
        logger.debug("Failed to use colored logs due to exception: %s", e)
        formatter = _logging.Formatter(fmt=_rc.log_format)
    for handler in logger.handlers:
        handler.setFormatter(formatter)


def init_logger(name: str = _rc.PROJECT_SLUG):
    logger = get_logger(name=name)
    logger.propagate = False
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logger.addHandler(_logging.StreamHandler())
    log_level = _rc.log_level
    log_level = log_level.upper() if isinstance(log_level, str) else log_level
    logger.setLevel(log_level)
    set_logger_formatter(logger)
    return logger



from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional

import torch as _torch

from tamm._adapters_v1 import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.common import AdapterMode as _AdapterMode
from tamm._adapters_v1.layer_adapters.common import (
    AdapterWithExtraInputs as _AdapterWithExtraInputs,
)
from tamm._adapters_v1.layer_adapters.common import _validate_adapters


class SoftMixingLoRA(_AdapterWithExtraInputs):
    """
    Applies a mixture of multiple LoRA adapters in parallel to the
    adapted layer, such that the output of the adapted layer is:

    output = adapted_layer(input) + lora_0(weight[:, 0] * inputs) +
        lora_1(weight[:, 1] * inputs) + lora_2(weight[:, 2] * inputs) + ...

    where weight is a m x n tensor where m is equal to the batch size of inputs
    and n is equal to the number of LoRA adapters being applied together.
    """

    extra_input_names: _List[str] = ["adapter_input_weights"]

    def __init__(
        self,
        adapters: _Dict[str, _LayerAdapter],
    ):
        super().__init__()
        _validate_adapters(adapters)
        self.register_child_adapters(adapters)
        self.adapters = self._child_adapters

    def forward(
        self,
        mode: _AdapterMode,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _Any,
        adapter_input_weights: _Optional[_torch.Tensor] = None,
    ):
        if mode is _AdapterMode.TRANSFORM_INPUTS:
            return self._transform_inputs(args=args, kwargs=kwargs)
        return self._transform_outputs(
            args=args,
            kwargs=kwargs,
            transformed_args=transformed_args,
            transformed_kwargs=transformed_kwargs,
            outputs=outputs,
            adapter_input_weights=adapter_input_weights,
        )

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _Any,
        adapter_input_weights: _Optional[_torch.Tensor] = None,
    ):
        if not isinstance(transformed_args, tuple):
            raise TypeError("SoftMixingLoRA expects a tuple for transformed_args")
        if len(transformed_args) != 1:
            raise ValueError("SoftMixingLoRA expects transformed_args of length 1")
        x = transformed_args[0]

        ad_idx = 0
        for adapter in self.adapters.values():
            input_weight = adapter_input_weights[:, ad_idx].reshape(
                -1, *([1] * (x.ndim - 1))
            )
            outputs = adapter(
                mode=_AdapterMode.TRANSFORM_OUTPUTS,
                args=args,
                kwargs=kwargs,
                transformed_args=(input_weight * x,),
                transformed_kwargs=transformed_kwargs,
                outputs=outputs,
            )
            ad_idx += 1
        return outputs



"""
Implements SentencePiece tokenizers for Apple Foundation Models.
"""

import functools as _functools
import os as _os
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Iterable as _Iterable
from typing import Optional as _Optional
from typing import Union as _Union

from tamm import _helpers
from tamm.tokenizers import text_processors as _text_processors
from tamm.tokenizers.sentencepiece import (
    SentencePieceTokenizer as _SentencePieceTokenizer,
)
from tamm.tokenizers.text_processors import TextProcessor as _TextProcessorBase
from tamm.tokenizers.text_processors import TextProcessorConfig as _TextProcessorConfig
from tamm.tokenizers.text_processors import TextProcessorMode as _TextProcessorMode
from tamm.utils import _pretrained


class AFMTokenizer(_SentencePieceTokenizer):
    """
    Tokenizer for AFM models, includes a default text processor that converts newline
    characters to a special token representation ``<n>``.
    """

    def __init__(
        self,
        vocab_path: _Optional[str] = None,
        user_defined_tokens: _Optional[_Iterable[str]] = None,
        extra_properties: _Dict[str, _Any] = None,
        text_processor: _Optional[
            _Union[str, _TextProcessorMode, _TextProcessorBase, _TextProcessorConfig]
        ] = "afm_text_find_and_replace",
    ):
        self._vocab_file_path = vocab_path

        if extra_properties is not None:
            for key, value in extra_properties.items():
                setattr(self, key, value)

        if isinstance(text_processor, str):
            text_processor = _helpers.get_enum_member_from_name(
                _TextProcessorMode, text_processor
            )
        if text_processor is _TextProcessorMode.AFM_TEXT_FIND_AND_REPLACE:
            text_processor = _text_processors.FindAndReplaceTextProcessor({"\n": "<n>"})
        elif isinstance(text_processor, _TextProcessorConfig):
            text_processor = text_processor()

        super().__init__(
            vocab_path=self.vocab_path,
            user_defined_tokens=user_defined_tokens,
            text_processor=text_processor,
        )

    @_functools.cached_property
    def vocab_path(self) -> str:
        """Returns the local path of the vocabulary file."""
        if _os.path.exists(self._vocab_file_path):
            return str(self._vocab_file_path)
        return _pretrained.fetch_file(remote_path=self._vocab_file_path).as_posix()



from torch import nn as _nn


class AttachMetadataPostHook:
    def __init__(self, metadata):
        self.metadata = metadata

    def __call__(self, model: _nn.Module) -> None:
        model.metadata = self.metadata



from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm.layers import sequential as _sequential
from tamm.layers import side_outputs as _side_outputs
from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.layers.transformer import token_metadata as _token_metadata
from tamm.layers.transformer.layer_sequence import common as _common
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder


class TransformerLayerSequence(
    _sequential.Sequential, _common.BaseTransformerLayerSequence
):
    """
    A generic sequence of transformer layers.  Each layer may be configured
    independently, and the architecture can vary between layers.  For example, layers
    0 and 1 could have different numbers of attention heads.

    Args:
        sequence: One or more transformer layers (or builders).
        output_hidden_states (:obj:`bool`): A flag for including outputs from each
            layer in the sequence output.  Defaults to ``False``.
        attention_types (:obj:`dict`, optional): A dictionary that
            maps layer indices to their secondary attention types.  For example
            if ``attention_types={3: "global", 7: "global"},
            then layers 3 and 7 receive the arguments
            ``attention_side_inputs=secondary_attention_side_inputs["global"]``
            rather than ``attention_side_inputs=attention_side_inputs``.
    """

    def __init__(
        self,
        *sequence: _ModuleOrBuilder,
        output_hidden_states: bool = False,
        attention_types: _Optional[_Dict[int, str]] = None,
    ):
        named_layers = {}
        side_input_keys = {}
        side_output_keys = {}

        for layer_idx, layer in enumerate(sequence):
            layer_name = f"layer_{layer_idx}"
            named_layers[layer_name] = layer
            side_input_keys[layer_name] = [
                "attention_side_inputs",
                (f"attention_side_inputs_{layer_idx}", "attention_side_inputs"),
                "cross_attention_side_inputs",
            ]
            if output_hidden_states:
                side_output_keys[layer_name] = "hidden_states"

        if attention_types is None:
            attention_types = {}
        self.attention_types = attention_types

        super().__init__(
            named_layers,
            side_input_keys=side_input_keys,
            side_output_keys=side_output_keys,
        )

    @property
    def num_layers(self):
        return len(self)

    def extra_repr(self):
        if self.attention_types:
            return f"attention_types={self.attention_types}"
        return ""

    # pylint: disable-next=arguments-differ,arguments-renamed
    def forward(
        self,
        hidden_states: _torch.Tensor,
        *,
        attention_side_inputs: _Dict[str, _Any],
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        secondary_attention_side_inputs: _Optional[_Dict[str, _Dict[str, _Any]]] = None,
        cross_attention_side_inputs: _Optional[_Dict[str, _Any]] = None,
        token_metadata: _token_metadata.TokenMetadata,
    ) -> _Union[_torch.Tensor, _side_outputs.OutputWithSideOutputs]:
        # pylint: disable=unused-argument
        attention_kwargs = self._get_attention_kwargs(
            attention_side_inputs=attention_side_inputs,
            secondary_attention_side_inputs=secondary_attention_side_inputs,
            kv_cache=kv_cache,
        )
        return super().forward(
            hidden_states,
            **attention_kwargs,
            cross_attention_side_inputs=cross_attention_side_inputs,
            token_metadata=token_metadata,
        )

    def _get_attention_kwargs(
        self, *, attention_side_inputs, secondary_attention_side_inputs, kv_cache
    ):
        if kv_cache is None and not self.attention_types:
            return {"attention_side_inputs": attention_side_inputs}

        prefix = "attention_side_inputs_"
        keys = [
            next(key for key, _ in side_input_keys_i if key.startswith(prefix))
            for side_input_keys_i in self.side_input_keys.values()
        ]

        result = {}
        for layer_idx, key in enumerate(keys):
            attention_type = self.attention_types.get(layer_idx, None)
            if attention_type is None:
                side_inputs_for_layer = attention_side_inputs
            else:
                side_inputs_for_layer = secondary_attention_side_inputs[attention_type]

            result[key] = dict(
                side_inputs_for_layer.items()
            )  # torch.compile friendly copy

            if kv_cache is not None:
                result[key]["kv_cache"] = kv_cache.at_layer(layer_idx)

        return result



"""
This module provides wrappers of tamm models for the Hugging Face
:mod:`transformers` ecosystem.
"""

# pylint: disable=wrong-import-position

import logging as _logging

_logger = _logging.getLogger(__name__)

try:
    import transformers as _hf_transformers
except Exception:
    _logger.error(
        "tamm.hf.transformers requires the Hugging Face transformers package, "
        "but tamm could not import this.  Please install transformers or check "
        "your transformers installation."
    )
    raise


from tamm import _plugin
from tamm.hf.transformers import modeling_utils
from tamm.hf.transformers.models.auto import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    register_with_auto_classes,
)
from tamm.hf.transformers.models.causal_lm import (
    TammCausalLMConfig,
    TammCausalLMForCausalLM,
    TammCausalLMModel,
)
from tamm.hf.transformers.models.sentencepiece import TammSentencePieceTokenizer

_plugin.execute_core_transformers_import_callbacks()



from torch import nn as _nn


class AttachConfigPostHook:
    def __init__(self, config):
        self.config = config

    def __call__(self, model: _nn.Module) -> None:
        model.config = self.config



import os as _os
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm.utils.optional_bool import OptionalBool

ModuleBuilder = _Callable[[], _nn.Module]
ModuleOrBuilder = _Union[_nn.Module, ModuleBuilder]
OptionalModuleOrBuilder = _Optional[ModuleOrBuilder]

DtypeOrString = _Union[_torch.dtype, str]
DeviceOrString = _Union[_torch.device, str]
OptionalDtypeOrString = _Optional[DtypeOrString]
OptionalDeviceOrString = _Optional[DeviceOrString]
LenientOptionalBool = _Optional[_Union[bool, OptionalBool]]
PathLike = _Union[str, bytes, _os.PathLike]

OptionalTensor = _Optional[_torch.Tensor]

StateDictType = _Dict[str, _torch.Tensor]



from typing import List as _List
from typing import Tuple as _Tuple

import torch as _torch

from tamm.layers.transformer.kv_cache.vanilla import VanillaKVCache as _VanillaKVCache
from tamm.layers.transformer.kv_cache.vanilla import (
    VanillaKVCacheView as _VanillaKVCacheView,
)


class SpeculativeKVCache(_VanillaKVCache):
    """
    An implementation of :class:`.BaseKVCache`, which handles speculative decoding
    scenarios.
    """

    def __init__(
        self,
        num_layers: int,
        batch_size: int,
        length: int,
        hidden_dim: int,
        *,
        device=None,
        dtype=None,
    ):
        super().__init__(
            num_layers, batch_size, length, hidden_dim, device=device, dtype=dtype
        )
        self.temporary_cache_view = None

    def create_temporary_cache_view(self, speculation_length: int):
        cache = _VanillaKVCache(
            self.num_layers,
            self.batch_size,
            speculation_length,
            self.hidden_dim,
            device=self.device,
            dtype=self.dtype,
        )
        self.temporary_cache_view = cache.up_to_index(speculation_length)

    def up_to_index(self, index: int) -> "SpeculativeKVCacheView":
        return SpeculativeKVCacheView(self, length=index)

    def to(self, *args, **kwargs) -> "SpeculativeKVCache":
        super().to(*args, **kwargs)
        self.temporary_cache_view.cache.to(*args, **kwargs)
        return self

    def merge(
        self, merge_length: _torch.Tensor, max_merge_length: int, previous_length: int
    ):
        if previous_length + max_merge_length > self.length:
            self.resize(self.length + self.length // 2)
        new_keys_and_values = self.temporary_cache_view.cache.keys_and_values[
            :, :, :max_merge_length
        ]

        segment_ids = self.temporary_cache_view.cache.segment_ids
        device = segment_ids.device
        # align sequences across batch dimension
        segment_ids[
            _torch.arange(segment_ids.shape[0], device=device)[:, None] >= merge_length
        ] = 0
        new_segment_ids = segment_ids[:max_merge_length]

        new_positions = self.temporary_cache_view.cache.positions[:max_merge_length]
        new_token_types = self.temporary_cache_view.cache.token_types[:max_merge_length]

        self.keys_and_values[
            :, :, previous_length : previous_length + max_merge_length
        ] = new_keys_and_values
        self.segment_ids[
            previous_length : previous_length + max_merge_length
        ] = new_segment_ids
        self.positions[
            previous_length : previous_length + max_merge_length
        ] = new_positions
        self.token_types[
            previous_length : previous_length + max_merge_length
        ] = new_token_types
        self.temporary_cache_view = None


class SpeculativeKVCacheView(_VanillaKVCacheView):
    """
    An implementation of :class:`._common.BaseKVCacheView`, which handles speculative
    decoding scenarios.

    Args:
        cache (:obj:`.SpeculativeKVCache`): A KV cache.
        length (:obj:`int`): The sequence length of the cache view.
    """

    def __init__(self, cache: SpeculativeKVCache, *, length: int):
        super().__init__(cache, length=length)
        self.cache: SpeculativeKVCache = cache

    def write_to_tail(
        self, key: _torch.Tensor, value: _torch.Tensor, *, layer_idx: int
    ) -> None:
        if self.cache.temporary_cache_view:
            self.cache.temporary_cache_view.write_to_tail(
                key, value, layer_idx=layer_idx
            )
        else:
            super().write_to_tail(key, value, layer_idx=layer_idx)

    def write_to_tail_vectorized(
        self, keys: _torch.Tensor, values: _torch.Tensor, *, layer_indices: _List[int]
    ) -> None:
        if self.cache.temporary_cache_view:
            self.cache.temporary_cache_view.write_to_tail_vectorized(
                keys, values, layer_indices=layer_indices
            )
        else:
            super().write_to_tail_vectorized(keys, values, layer_indices=layer_indices)

    def write_segment_ids_to_tail(self, segment_ids: _torch.Tensor) -> None:
        if self.cache.temporary_cache_view:
            self.cache.temporary_cache_view.write_segment_ids_to_tail(segment_ids)
        else:
            super().write_segment_ids_to_tail(segment_ids)

    def write_positions_to_tail(self, positions: _torch.Tensor) -> None:
        if self.cache.temporary_cache_view:
            self.cache.temporary_cache_view.write_positions_to_tail(positions)
        else:
            super().write_positions_to_tail(positions)

    def write_token_types_to_tail(self, token_types: _torch.Tensor) -> None:
        if self.cache.temporary_cache_view:
            self.cache.temporary_cache_view.write_token_types_to_tail(token_types)
        else:
            super().write_token_types_to_tail(token_types)

    def read(self, layer_idx: int) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        if self.cache.temporary_cache_view:
            orig_keys, orig_values = super().read(layer_idx)
            temp_keys, temp_values = self.cache.temporary_cache_view.read(layer_idx)
            keys = _torch.cat([orig_keys, temp_keys], dim=1)
            values = _torch.cat([orig_values, temp_values], dim=1)
            return keys, values
        return super().read(layer_idx)

    def read_vectorized(
        self, layer_indices: _List[int]
    ) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        if self.cache.temporary_cache_view:
            orig_ks, orig_vs = super().read_vectorized(layer_indices)
            temp_ks, temp_vs = self.cache.temporary_cache_view.read_vectorized(
                layer_indices
            )
            keys = _torch.cat([orig_ks, temp_ks], dim=1)
            values = _torch.cat([orig_vs, temp_vs], dim=1)
            return keys, values
        return super().read_vectorized(layer_indices)

    def read_segment_ids(self) -> _torch.Tensor:
        if self.cache.temporary_cache_view:
            orig_segment_ids = super().read_segment_ids()
            temp_segment_ids = self.cache.temporary_cache_view.read_segment_ids()
            return _torch.cat([orig_segment_ids, temp_segment_ids], dim=-1)
        return super().read_segment_ids()

    def read_positions(self) -> _torch.Tensor:
        if self.cache.temporary_cache_view:
            orig_positions = super().read_positions()
            temp_positions = self.cache.temporary_cache_view.read_positions()
            return _torch.cat([orig_positions, temp_positions], dim=-1)
        return super().read_positions()

    def read_token_types(self) -> _torch.Tensor:
        if self.cache.temporary_cache_view:
            orig_token_types = super().read_token_types()
            temp_token_types = self.cache.temporary_cache_view.read_token_types()
            return _torch.cat([orig_token_types, temp_token_types], dim=-1)
        return super().read_token_types()



"""
Provides activation layers.
"""

import abc as _abc
from typing import Callable as _Callable
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn
from torch.nn import functional as _F

from tamm.layers import functional as _tamm_F
from tamm.layers import lambda_layer as _lambda_layer
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.utils import registry as _registry_module

ActivationSpecType = _Union[str, list, tuple, _Callable, "Activation"]

ACTIVATION_REGISTRY = _registry_module.Registry("Activations")


def create_activation_layer(
    spec: _Union[ActivationSpecType, None]
) -> _Union["Activation", None]:
    """
    Factory function for creating activation layers from a variety of possible
    specification types.

    Args:
        spec: The specification for the activation.  If the arg is a valid key to the
            activation registry (such as ``"gelu"`` or ``"swi_glu"``), then the function
            creates the layer from the registry and returns it.  The ``spec`` arg can
            also be (1) an :obj:`Activation` instance, in which case it becomes the
            return object, (2) a callable that takes no arguments, in which case the
            returned layer is the result of the callable, or (3) the name of a function
            from ``torch.nn.functional``, in which case the function returns a
            :obj:`LambdaActivation` that wraps that function.

    Returns:
        An :obj:`Activation` instance corresponding to ``spec``.
    """
    if spec is None:
        return None
    if ACTIVATION_REGISTRY.is_valid_key(spec):
        return ACTIVATION_REGISTRY.create(spec)
    if isinstance(spec, Activation):
        return spec
    if callable(spec):
        return spec()
    is_torch_functional_name = isinstance(spec, str) and hasattr(_F, spec)
    if is_torch_functional_name:
        function = getattr(_F, spec)
        return LambdaActivation(function)
    return ACTIVATION_REGISTRY.create(spec)


class Activation(_nn.Module, _abc.ABC, _LayerMixin):
    """Abstract base class for activation layers."""

    def __init__(self):
        super().__init__()  # needed to remove nn.Module args from tamm Builder

    def __init_subclass__(cls, key=None, description=None, **kwargs):
        """Automatically register subclasses that provide a key."""
        super().__init_subclass__(**kwargs)
        if key is None:
            return
        ACTIVATION_REGISTRY.register(cls, key=key, description=description)

    @property
    def is_gated(self):
        """
        A flag with value ``True`` if the layer expects two tensors (with the same
        shape) as input and ``False`` otherwise.
        """
        return False

    @_abc.abstractmethod
    # pylint: disable-next=all
    def forward(self, input):
        """Subclasses implement forward"""


class LambdaActivation(_lambda_layer.Lambda, Activation):
    """
    Activation layer that wraps a callable function for :meth:`forward`.

    Args:
        function (:obj:`Callable`): A callable function that takes inputs to the forward
            function and returns the output of forward.
    """

    def __init__(self, function: _Callable):
        _lambda_layer.Lambda.__init__(self, function)
        Activation.__init__(self)


class GatedActivation(Activation):
    """
    Abstract base class for gated activation layers.  Gated activations receive a second
    input for gating.
    """

    @property
    def is_gated(self):
        return True

    @_abc.abstractmethod
    # pylint: disable-next=all
    def forward(self, gate_input, input):
        """Subclasses implement forward"""


class Tanh(_nn.Tanh, Activation, key="tanh"):
    """Tanh activation."""

    def __init__(self):
        _nn.Tanh.__init__(self)
        Activation.__init__(self)


class ReLU(_nn.ReLU, Activation, key="relu"):
    """ReLU activation."""

    def __init__(self):
        _nn.ReLU.__init__(self)
        Activation.__init__(self)


class GELU(_nn.GELU, Activation, key="gelu"):
    """GELU activation with tanh approximation by default."""

    def __init__(self, approximate="tanh"):
        _nn.GELU.__init__(self, approximate=approximate)
        Activation.__init__(self)


class SiLU(_nn.SiLU, Activation, key="silu"):
    """SiLU activation."""

    def __init__(self):
        _nn.SiLU.__init__(self)
        Activation.__init__(self)


class QuickGELU(Activation, key="quick_gelu"):
    """
    Quick GELU activation, which approximates GELU using ``x * sigmoid(1.702 * x)``.
    """

    def forward(self, input):  # pylint: disable=all
        return input * _torch.sigmoid(1.702 * input)


class GLU(GatedActivation, key="glu"):
    """Gated activation with sigmoid activation function (gated linear unit)."""

    # pylint: disable-next=all
    def forward(self, gate_input, input):
        return input * _torch.sigmoid(gate_input)


class SwiGLU(GatedActivation, key="swi_glu"):
    """GLU variant that uses Swish nonlinear function in place of sigmoid."""

    # pylint: disable-next=all
    def forward(self, gate_input, input):
        return _tamm_F.swi_glu(gate_input, input)


class GEGLU(GatedActivation, key="geglu"):
    """
    Implements `GeGLU < https://arxiv.org/abs/2002.05202 >`__,
    a variant of the gated linear unit activation function
    This implementation mirrors diffusers.models.activations.GEGLU.
    """

    # pylint: disable-next=all
    def forward(self, gate_input, input):
        return _tamm_F.geglu(gate_input, input)



import importlib as _importlib
import logging as _logging
from collections import defaultdict
from pathlib import Path as _Path
from typing import Any as _Any
from typing import List as _List
from typing import Mapping as _Mapping
from typing import Type as _Type

from tamm._plugin import discover_and_retrieve_plugins
from tamm.runtime_configuration import rc as _rc
from tamm.utils.uri._protocol import BaseURLHandler

_logger = _logging.getLogger(__name__)


def is_uri_or_posix(uri: str) -> bool:
    return _is_uri(uri) or _Path(uri).is_file()


def _is_uri(model_identifier: _Any):
    if not isinstance(model_identifier, str):
        return False
    return "://" in model_identifier


def _is_json_file(model_identifier: str):
    _path = _Path(model_identifier)
    return _path.suffix.lower() == ".json"


def _discover_plugin_uri_handlers():
    available_uri_handlers = {}
    for plugin, module_name in discover_and_retrieve_plugins().items():
        try:
            plugin_uri_module = _importlib.import_module(".utils.uri", module_name)
            new_handlers = {
                f"{module_name}:{handler.name}": handler
                for handler in plugin_uri_module.available_handlers
            }
            _logger.debug("Registering %s to available URI handlers", new_handlers)
            available_uri_handlers.update(new_handlers)
        except (ImportError, ModuleNotFoundError, AttributeError) as e:
            _logger.debug(
                "Plugin %s does not have any valid uri handlers: %s", plugin, e
            )
    return available_uri_handlers


class DiscoveredURIHandlers:
    def __init__(self, handlers: _Mapping[str, _Type["BaseURLHandler"]]):
        discovered_handlers: _Mapping[str, _List[str]] = defaultdict(list)
        for full_name in handlers.keys():
            base_name = full_name.split(":", maxsplit=2)[1]
            discovered_handlers[base_name].append(full_name)
        for base_name, full_names in discovered_handlers.items():
            if len(full_names) > 1:
                raise ValueError(
                    f"URI Handler named '{base_name}' is redefined multiple times in different plugins: {full_names}. "
                    "Uninstall one of the plugins or request plugin owners to resolve naming conflict"
                )

        self.base_name_to_namespaced_name = {
            k: v[0] for k, v in discovered_handlers.items()
        }
        self._handlers = handlers

    def __getitem__(self, item) -> _Type["BaseURLHandler"]:
        if item.lower() == "auto":
            try:
                return self._get_uri_handler(_rc.uri_handler)
            except KeyError as e:
                raise KeyError(
                    f"'auto' URI handler is set, but rc.uri_handler={_rc.uri_handler} does not exist"
                ) from e
        try:
            return self._get_uri_handler(item)
        except KeyError as e:
            raise KeyError(f"URI handler {item} does not exist") from e

    def _get_uri_handler(self, name: str) -> _Type["BaseURLHandler"]:
        try:
            return self._handlers[self.base_name_to_namespaced_name[name]]
        except KeyError:
            return self._handlers[name]

    @classmethod
    def discover(cls):
        return cls(_discover_plugin_uri_handlers())


# pylint: disable-next=invalid-name
def _URIHandler(*args, backend: str = "auto", **kwargs):
    """
    URI handler facade to determine which specific URI Handler to use based on the
    environment

    Args:
        backend: "auto", specific URI handler names by '<plugin_name>:<handler_name>' or '<handler_name>'
    """
    discovered_handlers = DiscoveredURIHandlers.discover()
    handler = discovered_handlers[backend]
    return handler(*args, **kwargs)



"""
layers.segmentation
^^^^^^^^^^^^^^^^^^^

This submodule implements layers for segmenting inputs.  Typically this involves
computing a padding mask.

.. autoclass:: tamm.layers.ConcatEmbeddingSegmentation
    :show-inheritance:
    :members: forward, create_builder_with_segmentation

.. autoclass:: tamm.layers.ConstantSegmentation
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.ConvEmbeddingPaddingTransform
    :show-inheritance:
    :members: forward

.. autoclass:: tamm.layers.TokensPaddingMask
    :show-inheritance:
    :members: forward

.. autofunction:: compute_right_padding_mask

.. autoclass:: tamm.layers.UnionSegmentation
    :show-inheritance:
    :members: forward
"""

from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm.layers import basic as _basic
from tamm.layers import common as _layers_common
from tamm.layers import functional as _tamm_F
from tamm.layers import sequential as _sequential
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class ConstantSegmentation(_nn.Module, _LayerMixin):
    """
    A layer that returns segment IDs with a fixed shape.

    Args:
        length (:obj:`int`): The size of the output in the sequence dimension.
        segment_id (:obj:`int`, optional): The segment ID for the outputs.  Defaults to
            ``1``.
    """

    def __init__(self, length: int, *, segment_id: int = 1):
        super().__init__()
        self.length = length
        self.segment_id = segment_id

    def extra_repr(self) -> str:
        return f"length={self.length}, segment_id={self.segment_id}"

    def forward(self, inputs: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            inputs (:obj:`torch.Tensor`): An input tensor where the first dimension
                is the batch size.

        Returns:
            A tensor with shape ``(batch_size, length)`` filled with the ``segment_id``
            value.
        """

        batch_size = inputs.size(0)
        return _torch.full(
            (batch_size, self.length), fill_value=self.segment_id, device=inputs.device
        )


class TokensPaddingMask(_nn.Module, _LayerMixin):
    """
    A layer that computes a padding mask from input token IDs.  Currently this layer
    only supports right padding, and left padding is treated as non-padding.  The
    outputs take value ``0`` for padding and ``1`` otherwise.

    .. see also:: :func:`.compute_right_padding_mask`

    Args:
        pad_token_id (:obj:`int`): The ID of the padding token.  Defaults to ``0``.
            This can also be ``None``, in which case the mask is entirely non-padding.
    """

    def __init__(self, pad_token_id: _Union[int, None] = 0):
        super().__init__()
        self.pad_token_id = pad_token_id

    def forward(self, input_ids: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            input_ids (:obj:`torch.Tensor`): A tensor of token IDs with shape
                ``(batch_size, seq_len)``.

        Returns:
            A right-padding mask.  For each input sequence, trailing tokens with value
            ``pad_token_id`` receive a ``0`` value for padding, and non-padding tokens
            receive a ``1``.
        """
        return compute_right_padding_mask(input_ids, pad_token_id=self.pad_token_id)

    def extra_repr(self):
        return f"pad_token_id={self.pad_token_id}"


def compute_right_padding_mask(
    input_ids: _torch.Tensor, *, pad_token_id: _Union[int, None] = 0
) -> _torch.Tensor:
    """
    Computes a right-padding mask.  If ``pad_token_id``, then for an input

    ::

        3, 2, 3, 0, 0
        0, 0, 0, 0, 0
        0, 5, 0, 5, 0

    we expect output

    ::

        1, 1, 1, 0, 0
        0, 0, 0, 0, 0
        1, 1, 1, 1, 0

    Args:
        input_ids (:obj:`torch.Tensor`): Sequences of token IDs with shape
            ``(batch_size, seq_len)``.
        pad_token_id (:obj:`int`): The ID of the padding token.  Defaults to ``0``.
            This can also be ``None``, in which case there is no padding.

    Returns:
        A right-padding mask with ``0`` values for padding and ``1`` for non-padding.
    """
    if pad_token_id is None:
        return _torch.ones_like(input_ids)

    padding_mask = input_ids.ne(pad_token_id)
    return (
        padding_mask.fliplr()
        .cumsum(dim=1, dtype=_torch.int32)
        .ne(0)
        .fliplr()
        .type(input_ids.dtype)
        # use cumsum here instead of cummax bc MPS does not support cummax in torch 2.0
    )


class UnionSegmentation(_basic.Union, _LayerMixin):
    """
    A layer for composing multiple child segmentation layers.

    Args:
        named_layers (:obj:`dict`): A dictionary that maps the name of each segmentation
            layer to a builder or :obj:`nn.Module`.
        flatten_outputs (:obj:`bool`): A flag for flattening outputs of the layer.
            Defaults to ``False``.
    """

    # pylint: disable=duplicate-code

    def __init__(
        self,
        named_layers: _Dict[str, _OptionalModuleOrBuilder],
        *,
        flatten_outputs: bool = False,
    ):
        super().__init__(named_layers)
        self.flatten_outputs = flatten_outputs

    def extra_repr(self) -> str:
        return f"flatten_outputs={self.flatten_outputs}"

    def forward(self, inputs: _Any) -> _Union[_torch.Tensor, _Dict[str, _Any]]:
        """
        Args:
            inputs: Inputs to the child segmentation layers.

        Returns:
            If ``flatten_outputs=False``, returns a :obj:`dict` that maps the name of
            each child layer to its output.  If ``flatten_outputs=True``, returns a
            single tensor with shape ``(batch_size, sequence_len)``.  This flat tensor
            results from flattening the sequence dimensions of each child output
            and then concatenating the outputs along the sequence dim in the order
            of the children.
        """
        result = super().forward(inputs)
        if self.flatten_outputs:
            return _tamm_F.maybe_flatten_sequence(result)
        return result


class ConvEmbeddingPaddingTransform(_nn.Module, _LayerMixin):
    """
    A layer for pooling padding values when using a :obj:`ConvEmbedding` layer.
    """

    def __init__(
        self,
        *,
        kernel_size: _Union[_Tuple[int, ...], int],
        stride: _Optional[_Union[_Tuple[int, ...], int]] = None,
        padding: _Optional[_Union[_Tuple[int, ...], int]] = 0,
        dilation: _Optional[_Union[_Tuple[int, ...], int]] = 1,
    ):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation

    def extra_repr(self):
        def check_all_equal(option, value):
            """Option may be either a tuple or a number"""
            try:
                for x in option:
                    if x != value:
                        return False
                return True
            except TypeError:
                return option == value

        extras = [f"kernel_size={self.kernel_size}", f"stride={self.stride}"]
        if not check_all_equal(self.padding, 0):
            extras.append(f"padding={self.padding}")
        if not check_all_equal(self.dilation, 1):
            extras.append(f"dilation={self.dilation}")
        return ", ".join(extras)

    @classmethod
    def create_builder_from_convolution_layer(
        cls, convolution
    ) -> _layers_common.LayerBuilder:
        """
        Creates a builder for a :obj:`ConvEmbeddingPaddingTransform` from a
        convolution layer.  This input must have ``kernel_size``, ``stride``,
        ``padding``, and ``dilation`` attributes.
        """
        return cls.Builder(
            kernel_size=convolution.kernel_size,
            stride=convolution.stride,
            padding=convolution.padding,
            dilation=convolution.dilation,
        )

    def forward(self, padding_mask: _torch.Tensor) -> _torch.Tensor:
        """
        Args:
            padding_mask (:obj:`torch.Tensor`): An ``n+1`` dimensional padding mask
                where ``n`` is the dimension of the convolution.  The first dimension
                is the batch dimension and the remaining dimensions must match the shape
                of the convolution's inputs along the conv dimensions (such as the
                height and width in the 2D case).  The mask should have ``0`` values for
                padding inputs and ``1`` otherwise.

        Returns:
            A max-pooled padding mask of shape ``(batch_size, seq_len)``.  Tokens have
            value ``1`` if they correspond to patches with at least one non-padding
            value.  Otherwise they have value ``0``.
        """
        input_dtype = padding_mask.dtype
        padding_mask = padding_mask.float()  # max_pool ops do not work with ints

        ndim = padding_mask.ndim - 1
        if ndim == 1:
            op = _torch.max_pool1d
        elif ndim == 2:
            op = _torch.max_pool2d
        elif ndim == 3:
            op = _torch.max_pool3d
        else:
            raise RuntimeError("Expected padding_mask to have 2, 3, or 4 dimensions")

        padding_mask = op(
            padding_mask,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
        )

        padding_mask = padding_mask.type(input_dtype)
        return padding_mask.reshape(padding_mask.size(0), -1)


class ConcatEmbeddingSegmentation(_nn.Module, _LayerMixin):
    """
    A segmentation layer to use with :obj:`.ConcatEmbedding` layers.  To account for
    concatenated embeddings, this layer concats extra segment IDs to the beginning
    and/or end of segment ID inputs.
    """

    def __init__(self, *, segment_id: int = 1, num_begin: int = 0, num_end: int = 0):
        super().__init__()
        self.num_begin = num_begin
        self.num_end = num_end
        self.segment_id = segment_id

    def extra_repr(self):
        return (
            f"segment_id={self.segment_id}, num_begin={self.num_begin}, "
            f"num_end={self.num_end}"
        )

    def forward(self, segment_ids: _torch.Tensor) -> _torch.Tensor:
        """
        Prepends and appends ``segment_id`` values to ``segment_ids`` along the sequence
        dimension.

        Args:
            segment_ids (:obj:`torch.Tensor`): Segment IDs with shape
                ``(batch size, seq len)``.
        """
        batch_size = segment_ids.size(0)
        pieces = [segment_ids]

        if self.num_begin > 0:
            begin_piece = _torch.full(
                size=(batch_size, self.num_begin),
                fill_value=self.segment_id,
                dtype=segment_ids.dtype,
                device=segment_ids.device,
            )
            pieces.insert(0, begin_piece)

        if self.num_end > 0:
            end_piece = _torch.full(
                size=(batch_size, self.num_end),
                fill_value=self.segment_id,
                dtype=segment_ids.dtype,
                device=segment_ids.device,
            )
            pieces.append(end_piece)

        if len(pieces) == 1:
            return pieces[0]
        return _torch.cat(pieces, dim=1)

    @classmethod
    def create_builder_with_segmentation(
        cls,
        segmentation_builder: _layers_common.LayerBuilder,
        *,
        segmentation_name: str = "segmentation",
        **kwargs,
    ):
        """
        Converts a builder for a segmentation layer into a builder for a sequence of
        two layers: the original segmentation and a :obj:`.ConcatEmbeddingSegmentation`.

        Args:
            segmentation_builder (:obj:`LayerBuilder`): A builder for a segmentation
                layer.
            segmentation_name (:obj:`str`, optional): A name for the segmentation layer.
                Defaults to ``"segmentation"``.
            kwargs: Keyword arguments for :class:`.ConcatEmbeddingSegmentation`.
        """
        builder = cls.Builder(**kwargs)
        layers = {
            segmentation_name: segmentation_builder,
            "concat_embedding_segmentation": builder,
        }
        return _sequential.Sequential.Builder(layers)



"""
Model Checksum
--------------

``tamm`` provides following utility functions for model checksum:

.. autofunction:: tamm.models.checksum.compare_checksum_with_source
.. autofunction:: tamm.models.checksum.compare_factory_checksum_with_source

Low-Level Functions
^^^^^^^^^^^^^^^^^^^
.. autofunction:: tamm.models.checksum.state_dict_checksum

"""

from tamm.models.checksum._checksum import state_dict_checksum
from tamm.models.checksum.api import (
    compare_checksum_with_source,
    compare_factory_checksum_with_source,
)

__all__ = [
    "state_dict_checksum",
    "compare_checksum_with_source",
    "compare_factory_checksum_with_source",
]



from tamm import tokenizers
from tamm._cli._listers._base import BaseLister
from tamm._cli._listers._helpers import (
    PrettyPrintListedEntities as _PrettyPrintListedEntities,
)


class TokenizerLister(BaseLister):
    name = "tokenizers"

    def print(self):
        _PrettyPrintListedEntities(tokenizers.list_tokenizers).print(
            include_descriptions=self.long,
            filter_deprecated=not self.show_deprecated,
            wide=self.wide,
        )



import abc as _abc
import contextlib as _contextlib
from typing import ContextManager


class _BaseContextHook(_abc.ABC):
    def __call__(self) -> "ContextManager":
        if self.is_null():
            return _contextlib.nullcontext()
        return self._get_context_manager()

    @_abc.abstractmethod
    def is_null(self) -> bool:
        ...

    @_abc.abstractmethod
    def _get_context_manager(self) -> "ContextManager":
        ...



"""
token_generators.base
^^^^^^^^^^^^^^^^^^^^^
"""

import abc as _abc
from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union
from typing import cast as _cast

import torch as _torch
from torch import nn as _nn

from tamm.utils.torch_utils import (
    torch_exportable_dataclass as _torch_exportable_dataclass,
)


@_torch_exportable_dataclass
class TokenGeneratorOutput:
    """
    Output generated from :py:class:`TokenGenerator`, consisting of token_ids and other
    fields specific to the generation algorithm.

    Attributes:
        token_ids (:py:class:`_torch.Tensor`): Token ids produced by a generator. Does not include ``input_ids``.
    """

    token_ids: _torch.Tensor


class TokenGenerator(_abc.ABC):
    """
    Base class for classes that generate tokens using an autoregressive decoder model.
    """

    def __init__(
        self,
        model: _nn.Module,
        eos_id: _Union[int, _torch.Tensor, _List[_torch.Tensor]],
        pad_id: _Union[int, _torch.Tensor],
    ):
        self.model = model
        self.eos_ids = self._cast_eos_ids_to_list(eos_id)
        self.pad_id = pad_id

    @_abc.abstractmethod
    def generate(
        self,
        input_ids: _torch.Tensor,
        max_new_tokens: int,
        seed: _Optional[int] = None,
    ) -> TokenGeneratorOutput:
        """
        Generate tokens using an autoregressive decoder model.
        """

    @staticmethod
    def _cast_eos_ids_to_list(
        eos_ids: _Union[int, _torch.Tensor, _List[_torch.Tensor], _List[int]]
    ) -> _List[_torch.Tensor]:
        # XXX: eos_ids may be one of two things with different semantics;
        #
        # a) a single torch.Tensor of various single-shot tokens, each of which terminates decoding
        # b) a list of 1D torch.Tensors, any of which must be found in its entirety in a decode in order to
        # terminate decoding
        #
        # we determine which we have been passed here and cast it into the latter form
        if isinstance(eos_ids, int):
            return [_torch.tensor([eos_ids])]
        if isinstance(eos_ids, _torch.Tensor):
            return list(eos_ids.flatten())
        if isinstance(eos_ids, list) and isinstance(eos_ids[0], int):
            return [_torch.tensor(token, dtype=_torch.int) for token in eos_ids]
        return _cast(_List[_torch.Tensor], eos_ids)



import abc as _abc
import math as _math
from typing import Optional as _Optional

import torch as _torch

from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class MixtureOfExpertsCapacityConstraintOutput:
    """
    A result dataclass for :obj:`MixtureOfExpertsCapacityConstraint` layers.

    Args:
        assignment_indicators (:obj:`torch.Tensor`): An updated indicator
            tensor with shape
            ``(*track_shape, *batch_shape, num_experts_per_token, num_experts)``.
            A value of ``1`` indicates a token assigned to an expert (and not
            dropped).
        capacity_per_expert (:obj:`int`): The total number of tokens to dispatch
            per expert.
    """

    assignment_indicators: _torch.Tensor
    capacity_per_expert: int


class MixtureOfExpertsCapacityConstraint(_torch.nn.Module, _abc.ABC):
    """
    A base class for capacity constraint layers in
    :obj:`CapacityConstrainedMixtureOfExpertsDispatcher` layers.
    """

    @_abc.abstractmethod
    def forward(
        self, assignment_indicators: _torch.Tensor, num_track_dims: int
    ) -> MixtureOfExpertsCapacityConstraintOutput:
        """
        Determines which tokens to drop as well as the token capacity of
        experts.

        Args:
            assignment_indicators (:obj:`torch.Tensor`): An indicator tensor
                with shape
                ``(*track_shape, *batch_shape, num_experts_per_token, num_experts)``.
                A value of ``1`` indicates a token assigned to an expert.
            num_track_dims (:obj:`int`): The number of track dimensions in
                ``assignment_indicators``.

        Returns:
            A :obj:`MixtureOfExpertsCapacityConstraintOutput`.
        """


class VanillaMixtureOfExpertsCapacityConstraint(
    MixtureOfExpertsCapacityConstraint, _LayerMixin
):
    """
    A basic capacity constraint that determines the capacity using a
    capacity factor.  This layer prioritizes tokens first based on the
    ``num_experts_per_token`` index, then the sequence index, and finally the
    batch index.  Tokens with larger indices are more likely to drop.

    Args:
        capacity_factor (:obj:`float`): The amount of expert capacity relative
            to the number of tokens.  A value of ``1`` results in the same
            capacity as the number of tokens, but it likely results in token
            dropping due to load imbalances.  Larger values result in more
            computation but less token dropping.
        capacity_factor_eval (:obj:`float`, optional): An optional capacity value
            for eval mode.  Defaults to ``capacity_factor``.
    """

    def __init__(
        self,
        *,
        capacity_factor: float,
        capacity_factor_eval: _Optional[float] = None,
    ):
        super().__init__()

        if capacity_factor_eval is None:
            capacity_factor_eval = capacity_factor
        self.capacity_factor = capacity_factor
        self.capacity_factor_eval = capacity_factor_eval

    def extra_repr(self):
        return (
            f"capacity_factor={self.capacity_factor}, "
            f"capacity_factor_eval={self.capacity_factor_eval}"
        )

    def forward(
        self, assignment_indicators: _torch.Tensor, num_track_dims: int
    ) -> MixtureOfExpertsCapacityConstraintOutput:
        # assignment_indicators has shape (T, B, S, K, E)

        assignment_indicators = assignment_indicators.swapdims(
            num_track_dims,
            -2
            # this prioritizes larger K indices for token dropping and avoids
            # prioritizing based on the batch index
        )  # (T, K, S, B, E)

        tokens_shape = assignment_indicators.shape[num_track_dims:-1]  # (K, S, B)
        assignment_indicators = assignment_indicators.flatten(
            start_dim=num_track_dims, end_dim=-2
        )  # (T, K*S*B, E)

        capacity_per_expert = self._get_capacity_per_expert(
            num_tokens=tokens_shape.numel(), num_experts=assignment_indicators.size(-1)
        )

        capacity_indices = _torch.cumsum(assignment_indicators, dim=-2)  # (T, K*S*B, E)
        capacity_indices *= assignment_indicators
        capacity_indices *= capacity_indices <= capacity_per_expert
        updated_indicators = capacity_indices.not_equal_(0)  # (T, K*S*B, E)

        updated_indicators = updated_indicators.unflatten(
            num_track_dims, tokens_shape
        )  # (T, K, S, B, E)

        updated_indicators = updated_indicators.swapdims(
            num_track_dims, -2
        )  # (T, B, S, K, E)

        return MixtureOfExpertsCapacityConstraintOutput(
            assignment_indicators=updated_indicators,
            capacity_per_expert=capacity_per_expert,
        )

    def _get_capacity_per_expert(self, *, num_tokens: int, num_experts: int):
        if self.training:
            capacity_factor = self.capacity_factor
        else:
            capacity_factor = self.capacity_factor_eval
        return _math.ceil(capacity_factor * num_tokens / num_experts)



"""
tamm.runtime_configuration
--------------------------

.. data:: tamm.rc

    The global :obj:`.RuntimeConfiguration` instance.

.. autoclass:: tamm.runtime_configuration.RuntimeConfiguration
    :members:

.. autoclass:: tamm.runtime_configuration.PluginExtrasEntrypointNames
    :members:
"""

import dataclasses as _dataclasses
import functools as _functools
import os as _os
import pathlib as _pathlib
from typing import Any as _Any
from typing import Callable as _Callable
from typing import List as _List
from typing import Optional as _Optional


def _to_str(value: _Any) -> str:
    """Small helper for formatting values in docstrings."""
    if isinstance(value, str):
        value = f'"{value}"'
    return f"``{value}``"


class RCConstant:
    """
    A helper class for creating constant properties in :class:`.RuntimeConfiguration`.
    The class is expected to be used through the :method:`create_property` class method.

    Args:
        value: The constant value of the property.
        doc: A description for the property's docstring.
    """

    def __init__(self, *, value: _Any, doc: str):
        self.value = value
        if not doc.endswith("."):
            doc = doc + "."
        self.doc = doc

    @classmethod
    def create_property(cls, *, value: _Any, doc: str) -> property:
        obj = cls(value=value, doc=doc)
        return property(fget=obj.create_getter(), doc=obj.create_doc())

    def create_getter(self):
        def getter(_):
            return self.value

        return getter

    def create_doc(self):
        return f"{self.doc} This is a constant with value {_to_str(self.value)}."


class RCVariable:
    """
    A helper class for creating variable properties in :class:`.RuntimeConfiguration`.
    These are configuration options that can modified, either through environment variable
    or by setting them on the :obj:`.RuntimeConfiguration` instance. The class is expected
    to be used through the :method:`create_property` class method.

    Args:
        name: The name of the property.
        default: The property's default value.
        doc: A description for the property's docstring.
        parse_hook (:obj:`callable`, optional): An optional function for parsing the
            variable's value when it is specified as an environment variable.
        choices (:obj:`list`, optional): An optional list of possible values for the
            variable.  If specified, the property raises a :obj:`ValueError` if the
            user sets it to a value not included in the list.
        runtime_changeable (:obj:`bool`, optional): A flag that controls whether the
            variable can be modified after creation.  If ``False``, the variable can
            only be specified via environment variable.
    """

    def __init__(
        self,
        *,
        name: str,
        default: _Any,
        doc: str,
        parse_hook: _Optional[_Callable[[_Any], _Any]] = None,
        choices: _Optional[_List[_Any]] = None,
        runtime_changeable: bool = True,
    ):
        self.name = name
        self.default = default
        if not doc.endswith("."):
            doc = doc + "."
        self.doc = doc
        self.parse_hook = parse_hook
        self.choices = choices
        self.is_runtime_changeable = runtime_changeable

    @classmethod
    def create_property(
        cls,
        *,
        name: str,
        default: _Any,
        doc: str,
        parse_hook: _Optional[_Callable[[_Any], _Any]] = None,
        choices: _Optional[_List[_Any]] = None,
        runtime_changeable: bool = True,
    ):
        obj = cls(
            name=name,
            default=default,
            doc=doc,
            parse_hook=parse_hook,
            choices=choices,
            runtime_changeable=runtime_changeable,
        )
        return property(
            fget=obj.create_getter(),
            fset=obj.create_setter(),
            fdel=obj.create_deleter(),
            doc=obj.create_doc(),
        )

    @property
    def env_var_name(self):
        return f"TAMM_{self.name}".upper()

    @property
    def attr_name(self):
        return f"_{self.name}"

    @property
    def updated_default(self):
        if self.env_var_name not in _os.environ:
            return self.default
        result = _os.environ[self.env_var_name]
        if self.parse_hook is not None:
            return self.parse_hook(result)
        return result

    def _validate_choice(self, value):
        if self.choices is None:
            return
        if value not in self.choices:
            raise ValueError(f"Value '{value}' is not a choice from {self.choices}")

    def create_getter(self):
        def getter(self2):
            if not hasattr(self2, self.attr_name):
                value = self.updated_default
                self._validate_choice(value)
                if not self.is_runtime_changeable:
                    setattr(
                        self2,
                        self.attr_name,
                        value
                        # recording the value of updated_default ensures the
                        # value does not change if the env is later modified
                    )
                return value
            return getattr(self2, self.attr_name)

        return getter

    def create_setter(self):
        def setter(self2, new_value):
            if not self.is_runtime_changeable:
                raise NotRuntimeChangeableException(
                    name=self.name, env_var_name=self.env_var_name
                )
            self._validate_choice(new_value)
            setattr(self2, self.attr_name, new_value)
            self.setter_post_hook(new_value)

        return setter

    def create_deleter(self):
        def deleter(self2):
            if hasattr(self2, self.attr_name):
                delattr(self2, self.attr_name)
            self.setter_post_hook(self.updated_default)

        return deleter

    def setter_post_hook(self, value: _Any) -> None:
        """
        A hook that runs whenever the property's value changes.  Subclasses
        can override this method to introduce callback behavior.
        """

    def create_doc(self):
        full_doc = [self.doc]
        if self.is_runtime_changeable:
            full_doc.append(
                f"Configure using :attr:`tammm.rc` or the ``{self.env_var_name}`` env var."
            )
        else:
            full_doc.append(
                f"Only configurable using the ``{self.env_var_name}`` env var."
            )

        if self.choices is not None:
            choices = [_to_str(c) for c in self.choices]
            if len(choices) == 1:
                choices_str = f"{choices[0]} only"
            elif len(choices) == 2:
                choices_str = f"{choices[0]} and {choices[1]}"
            else:
                choices_strings = [f"{c}" for c in choices]
                choices_strings[-1] = f"and {choices_strings[-1]}"
                choices_str = ", ".join(choices_strings)
            full_doc.append(f"Choices are {choices_str}.")

        full_doc.append(f"Defaults to {_to_str(self.default)}.")
        return " ".join(full_doc)


class LogLevelRCVariable(RCVariable):
    def setter_post_hook(self, value: _Any) -> None:
        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from tamm import _logger

        logger = _logger.get_logger()
        logger.setLevel(value)


class LogFormatRCVariable(RCVariable):
    def setter_post_hook(self, value: _Any) -> None:
        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from tamm import _logger

        logger = _logger.get_logger()
        _logger.set_logger_formatter(logger)


def _parse_bool(value: str) -> bool:
    try:
        return bool(int(value))
    except ValueError as exc:
        if value.lower() in {"true", "yes", "y", "t"}:
            return True
        if value.lower() in {"false", "no", "n", "f"}:
            return False
        raise ValueError(f"Cannot parse value '{value}' to boolean") from exc


class NotRuntimeChangeableException(Exception):
    """
    When :class:`SessionVariable`().value is assigned, this exception will be raised
    if :attr:`runtime_changeable` is `False`
    """

    def __init__(self, name: str, env_var_name: str):
        super().__init__(
            f"Session variable '{name}' is not writable during runtime, "
            f"restart the session with environment variable "
            f"'{env_var_name}' instead."
        )


@_dataclasses.dataclass(frozen=True)
class PluginExtrasEntrypointNames:
    """A frozen dataclass for organizing entrypoint name constants for plugin extras."""

    core_import_callback: str = "core-import-callback"
    """The entrypoint name for callbacks that execute after |tamm| import."""

    core_transformers_import_callback: str = "core-transformers-import-callback"
    """The entrypoint name for callbacks that execute after :mod:`tamm.hf.transformers` import."""

    cli: str = "cli"
    """The entrypoint name for submodules that define |tamm| CLI commands."""

    tokenizer_registry: str = "tokenizer-registry"
    """The entrypoint name for plugins that define **a factory of** tokenizer registry."""

    preprocessor_registry: str = "preprocessor-registry"
    """The entrypoint name for plugins that define **a factory of** preprocessor registry."""


class RuntimeConfigurationBase:
    def reset(self, *names: str) -> None:
        """
        Reset RC variables to their defaults.

        Args:
            *names (:obj:`str`, optional):  The name of the property (or properties) to reset.
                If empty, the function resets all properties.
        """
        if len(names) == 0:
            names = tuple(
                key
                for key, val in vars(self.__class__).items()
                if isinstance(val, property)
            )
        for name in names:
            try:
                delattr(self, name)
            except AttributeError:
                pass


class RuntimeConfiguration(RuntimeConfigurationBase):
    """A class that organizes |tamm| global constants and variables."""

    flash_attn_max_num_segments = RCVariable.create_property(
        name="flash_attn_max_num_segments",
        default=8,
        doc="Max number of segments for flash attention (must be a multiple of 8)",
    )

    log_format = LogFormatRCVariable.create_property(
        name="log_format",
        default="%(asctime)s:%(name)s:%(lineno)s:%(levelname)s: %(message)s",
        doc="Log format of the root |tamm| :obj:`logging.Logger`",
    )

    log_level = LogLevelRCVariable.create_property(
        name="log_level",
        default="INFO",
        doc="Log level of the root |tamm| :obj:`logging.Logger`",
    )

    uri_handler = RCVariable.create_property(
        default="boto3",
        name="uri_handler",
        doc="Default URI handler",
        parse_hook=lambda x: x.lower(),
    )

    user_dir = RCVariable.create_property(
        default=str(_pathlib.Path().home() / ".tamm"),
        name="user_dir",
        doc="A local directory for caching downloads and storing tmp files",
    )

    PROJECT_SLUG = RCConstant.create_property(value="tamm", doc="Name of this project")

    BASE_ADAPTER_GROUP_NAME = RCConstant.create_property(
        value="base_adapter",
        doc="Name of the default adapter if model is adapted",
    )

    MARKER_ATTRIBUTE = RCConstant.create_property(
        value="_tamm_marker",
        doc="nn.Module attribute name to keep track of extra markers",
    )

    PLUGINS_ENTRYPOINT_GROUP = RCConstant.create_property(
        value="tamm.plugins", doc="The entrypoint group for |tamm| plugins"
    )

    PLUGINS_EXTRAS_ENTRYPOINT_GROUP = RCConstant.create_property(
        value="tamm.plugins.extras", doc="The entrypoint group for |tamm| plugin extras"
    )

    PLUGINS_EXTRAS_ENTRYPOINT_NAMES = RCConstant.create_property(
        value=PluginExtrasEntrypointNames(),
        doc="The :obj:`.PluginExtrasEntrypointNames` for |tamm| plugin extras",
    )

    adapters_implementation = RCVariable.create_property(
        name="adapters_implementation",
        default="v1",
        doc="Implementation of :mod:`tamm.adapters`",
        choices=["v1"],
        parse_hook=lambda x: x.lower(),
        runtime_changeable=False,
    )


@_functools.cache
def _get_rc() -> RuntimeConfiguration:
    return RuntimeConfiguration()


rc = _get_rc()



"""
This submodule contains |tamm| versions of layers from :mod:`torch.nn`.  These layers
have the same behavior as their :mod:`torch` counterparts.  The difference is that the
|tamm| versions also have a :class:`LayerBuilder` attached as the ``Builder`` class
attribute.
"""

import torch.nn as _nn

from tamm.layers.common import LayerMixin as _LayerMixin


class MaxPool2d(_nn.MaxPool2d, _LayerMixin):
    """A |tamm| version of :class:`torch.nn.MaxPool2d`."""


class AdaptiveAvgPool2d(_nn.AdaptiveAvgPool2d, _LayerMixin):
    """A |tamm| version of :class:`torch.nn.AdaptiveAvgPool2d`."""


class AvgPool2d(_nn.AvgPool2d, _LayerMixin):
    """A |tamm| version of :class:`torch.nn.AvgPool2d`."""


class BatchNorm2d(_nn.BatchNorm2d, _LayerMixin):
    """A |tamm| version of :class:`torch.nn.BatchNorm2d`."""


class GroupNorm(_nn.GroupNorm, _LayerMixin):
    """A |tamm| version of :class:`torch.nn.GroupNorm`."""


class Softmax(_nn.Softmax, _LayerMixin):
    """A |tamm| version of :class:`torch.nn.Softmax`."""



"""
Architecture optimizers
=======================

Architecture optimizers are high-level objects that attach layers from
:mod:`tamm.ao.layers` to |tamm| models.  They are also configurable and
JSON-serializable.


.. autoclass:: tamm.ao.arch_optimizers.ArchOptimizer
    :members:

.. autoclass:: tamm.ao.arch_optimizers.KVQuantArchOptimizer
    :members:
"""

from tamm.ao.arch_optimizers.common import ArchOptimizer
from tamm.ao.arch_optimizers.kv_quant import KVQuantArchOptimizer

__all__ = ["ArchOptimizer", "KVQuantArchOptimizer"]



"""
.. _tamm_tokenizers:

tamm.tokenizers
===============

Create tokenizer
----------------
.. autofunction:: tamm.create_tokenizer

.. autofunction:: tamm.create_tokenizer_config

.. autofunction:: tamm.create_tokenizer_from_model_id

.. autofunction:: tamm.create_tokenizer_from_tokenizer_id

Tokenizers Base Classes
-----------------------

.. autoclass:: tamm.tokenizers.common.Tokenizer
   :show-inheritance:
   :members:
   :member-order: bysource

.. autoclass:: tamm.tokenizers.sentencepiece.SentencePieceTokenizer
   :show-inheritance:
   :members:
   :member-order: bysource

.. autoclass:: tamm.tokenizers.afm.AFMTokenizer
   :show-inheritance:
   :members:

Text Processors
---------------

.. autoclass:: tamm.tokenizers.text_processors.TextProcessor
   :show-inheritance:
   :members:
   :member-order: bysource

.. autoclass:: tamm.tokenizers.text_processors.FindAndReplaceTextProcessor
   :show-inheritance:
   :members:
   :member-order: bysource

.. automodule:: tamm.tokenizers.registry
"""

from tamm.tokenizers import afm, common, sentencepiece, text_processors
from tamm.tokenizers.api import (
    TokenizerSpecType,
    create_tokenizer,
    create_tokenizer_config,
    create_tokenizer_from_model_id,
    create_tokenizer_from_tokenizer_id,
    list_tokenizers,
)



# pylint: disable=unused-argument

from typing import Any as _Any
from typing import Callable as _Callable
from typing import Dict as _Dict

import torch.nn as _nn
from torch.utils.hooks import RemovableHandle as _RemovableHandle


class SharedContext:
    """
    An object for enabling context sharing between modules.

    Provides pre and post forward hooks which create a shared context
    dictionary, populate it with data, consume the data and delete the
    shared dictionary.

    In a typical use case, :py:meth:`register_context_creation_hook` and
    :py:meth:`register_context_deletion_hook` are called on ``ModuleA``
    and :py:meth:`register_context_population_hook` and
    :py:meth:`register_context_retrieval_hook` are called on its submodules.
    """

    def __init__(self):
        self._context = None

    def _create_context(self):
        """
        Create shared context dictionary.
        """
        self._context = {}

    def _delete_context(self):
        """
        Delete shared context dictionary.
        """
        del self._context
        self._context = None

    def register_context_creation_hook(
        self, module: _nn.Module, *input_keys: str
    ) -> _RemovableHandle:
        """
        Register a pre-forward hook on the module which creates
        the shared context dictionary.

        Args:
            module (:py:class:`_nn.Module`): Module on which hook is registered
        """

        def context_creation_pre_forward_hook(
            mod: _nn.Module, args: _Any, kwargs: _Any
        ):
            self._create_context()

        return module.register_forward_pre_hook(
            context_creation_pre_forward_hook, prepend=True, with_kwargs=True
        )

    def register_context_population_hook(
        self, module: _nn.Module, hook_fn: _Callable, prepend: bool = False
    ) -> _RemovableHandle:
        """
        Register a pre-forward hook on a module which populates the shared context
        dictionary with data that is consumed by other modules.

        Args:
            module (:py:class:`_nn.Module`): Module on which hook is registered
            hook_fn (:obj:`callable`): A callable with the signature
                ``hook(module, context, args, kwargs)``, where context is the shared
                context dictionary, and args and kwargs are inputs to the forward method
                of the module. This callable populates the context dictionary
                with data and optionally changes the args and kwargs which are
                passed to module.
            prepend (:obj:`bool`):  If true, the provided hook will be fired
                before all existing forward_pre hooks on module. Otherwise, the
                provided hook will be fired after all existing forward_pre hooks.
                Defaults to ``False``.
        """

        def context_population_pre_forward_hook(
            mod: _nn.Module, args: _Any, kwargs: _Any
        ):
            return hook_fn(mod, self._context, args, kwargs)

        return module.register_forward_pre_hook(
            context_population_pre_forward_hook, prepend=prepend, with_kwargs=True
        )

    @staticmethod
    def get_input_forwarding_hook(*input_keys: str) -> _Callable:
        """
        Returns a hook with the signature ``hook(module, context, args, kwargs)``
        which extracts inputs identified by ``input_keys`` from the kwargs passed to
        module and adds them to the context dictionary.

        This hook can be passed to :py:meth:`register_context_population_hook`
        for enabling a module to forward inputs passed to it to other submodules.

        Args:
            *input_keys: Variable number of names of extra inputs which are extracted
                from kwargs passed to module
        """

        def input_forwarding_hook(
            mod: _nn.Module, context: _Dict[str, _Any], args: _Any, kwargs: _Any
        ):
            new_kwargs = {}
            for key, val in kwargs.items():
                if key in input_keys:
                    context[key] = val
                else:
                    new_kwargs[key] = val
            return args, new_kwargs

        return input_forwarding_hook

    def register_context_deletion_hook(self, module: _nn.Module) -> _RemovableHandle:
        """
        Register a forward hook on the module which deletes the shared context
        dictionary, created by the context creation hook.

        This hook should be registered on a module which is guaranteed to be executed
        after all the modules which depend on the shared context for data
        have been executed.

        For a typical use case, both context creation hook and deletion
        hook are installed on a top level module, which forwards some extra
        inputs passed to it to some of its submodules, which expect these
        extra inputs.

        Args:
            module (:py:class:`_nn.Module`): Module on which hook is registered
        """

        def context_deletion_forward_hook(mod: _nn.Module, args: _Any, output: _Any):
            self._delete_context()

        return module.register_forward_hook(context_deletion_forward_hook, prepend=True)

    def register_context_retrieval_hook(
        self, module: _nn.Module, *input_keys: str
    ) -> _RemovableHandle:
        """
        Register hook on a submodule which retrieves data, identified by
        ``input_keys`` from the shared context dictionary.

        Args:
            module (:py:class:`_nn.Module`): Module on which hooks are registered
            *input_keys: Variable number of names of extra inputs which are retrieved
                from the shared context dictionary
        """

        def context_retrieval_hook(mod: _nn.Module, args: _Any, kwargs: _Any):
            for key in input_keys:
                if key in self._context:
                    kwargs[key] = self._context[key]
            return args, kwargs

        return module.register_forward_pre_hook(
            context_retrieval_hook, prepend=True, with_kwargs=True
        )



"""
transformer.layer
^^^^^^^^^^^^^^^^^

This module implements a transformer layer.

.. autoclass:: tamm.layers.TransformerLayer
    :members:
"""

from tamm.layers import sequential as _sequential
from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class TransformerLayer(_sequential.Sequential, _LayerMixin):
    """
    A sequence of (self) attention, optional cross-attention, and finally feed
    forward. The :meth:`forward` method is equivalent to:

    .. code-block:: python

        def forward(
            self, inputs, attention_side_inputs, cross_attention_side_inputs
        ):
            x = self.attention(inputs, **attention_side_inputs)

            if self.cross_attention is not None:
                x = self.cross_attention(hidden_states, **cross_attention_side_inputs)

            return self.feed_forward(x)

    Typically the layer takes inputs of shape
    ``(batch_size, sequence_length, input_dim)`` and returns outputs of the same shape.

    Args:
        attention: A self-attention layer.
        cross_attention: An optional cross-attention layer.
        feed_forward: A feed-forward layer.
    """

    def __init__(
        self,
        *,
        attention: _ModuleOrBuilder,
        cross_attention: _OptionalModuleOrBuilder = None,
        feed_forward: _ModuleOrBuilder,
    ):
        named_layers = {}
        side_input_keys = {}

        named_layers["attention"] = attention
        side_input_keys["attention"] = [
            # individual args only included for backward compatibility:
            "attention_mask",
            "flash_attention_options",
            "kv_cache",
            "query_rope_coefficients",
            "key_rope_coefficients",
            # keyword arguments:
            ("attention_side_inputs", "**kwargs"),
        ]

        if cross_attention:
            named_layers["cross_attention"] = cross_attention
            side_input_keys["cross_attention"] = [
                # keyword arguments:
                ("cross_attention_side_inputs", "**kwargs"),
            ]

        named_layers["feed_forward"] = feed_forward

        super().__init__(named_layers, side_input_keys=side_input_keys)



import contextlib as _contextlib
from contextvars import ContextVar
from typing import Union

from tamm.utils.optional_bool import OptionalBool

pretrained = ContextVar("pretrained", default=OptionalBool.FALSE)


def get_default_pretrained_flag():
    """
    ``.create_builder()`` of each configurable module will use the return value to
    resolve whether to load pretrained weights, *if* invoked without an explict
    Boolean value (i.e., ``None``, or ``OptionalBool.NOTSET``).

    Default value is ``False``
    """
    return pretrained.get()


def set_default_pretrained_flag(value: Union[bool, OptionalBool]):
    """
    Sets |tamm| global ``pretrained`` value

    Args:
        value: True, False, OptionalBool.[TRUE, FALSE]
    """
    value = OptionalBool(value)
    if value not in {OptionalBool.TRUE, OptionalBool.FALSE}:
        raise ValueError(
            "pretrained can only be either 'OptionalBool.TRUE' or 'OptionalBool.FALSE'"
        )
    return pretrained.set(value)


@_contextlib.contextmanager
def pretrained_flag_context(value):
    """
    Context of |tamm| global ``pretrained``

    Example:

    .. code-block:: python

        with pretrained_flag_context(True):
            tamm.create_model(...) # pretrained=True not needed

    """
    token = set_default_pretrained_flag(value)
    try:
        yield
    finally:
        pretrained.reset(token)



"""
adapters.layer_annotations
==========================

.. autoclass:: tamm.adapters.layer_annotations.FusedLinearProjection
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.AttentionInputTransformFused
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.LinearProjection
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.AttentionInputTransformQuery
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.AttentionInputTransformKey
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.AttentionInputTransformValue
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.AttentionOutputTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.FeedForwardHiddenTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.FeedForwardOutputTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedFusedLinearProjection
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedAttentionInputTransformFused
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedLinearProjection
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedAttentionInputTransformQuery
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedAttentionInputTransformKey
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedAttentionInputTransformValue
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedAttentionOutputTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedFeedForwardHiddenTransform
    :show-inheritance:
    :members:

.. autoclass:: tamm.adapters.layer_annotations.BatchedFeedForwardOutputTransform
    :show-inheritance:
    :members:
"""

import dataclasses as _dataclasses
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Type as _Type
from typing import Union as _Union

import torch.nn as _nn

from tamm.utils import registry as _registry

_LAYER_ANNOTATION_FIELD = "_layer_annotation_type"

_LAYER_ANNOTATION_TYPE_REGISTRY = _registry.Registry("Layer annotation types")


class LayerAnnotation:
    """
    An annotation data class which defines a contract between
    an adapter :py:class:`BaseAdapter` and a module which can be
    adapted using that adapter. This contract typically consists of
    names of fields and parameters of the module which are recognized
    by the adapter.

    This contract also enables a :py:class:`ModelAdapter`
    to correctly recognize different submodules of same type in the model
    and adapt them according to their annotations.

    Any new layer type can be added by subclassing :py:class:`LayerAnnotation`,
    which also automatically enables it to be specified using a
    string corresponding to its class name.
    """

    def __init_subclass__(cls):
        _dataclasses.dataclass(cls)
        _LAYER_ANNOTATION_TYPE_REGISTRY.register(cls, cls.__name__)


LayerAnnotationType = _Union[
    _Tuple[_Union[_Type[LayerAnnotation], str], _Optional[_Dict[str, str]]],
    _Tuple[_Union[_Type[LayerAnnotation], str]],
]


class FusedLinearProjection(LayerAnnotation):
    """
    Annotates a fused linear projection layer, which applies
    the following transformation to inputs `X`:

    ``output = split(W @ X + b, output_dims, dim=-1)``
    """

    input_dim_field: str
    output_dims_field: str
    weight_param_name: str


class BatchedFusedLinearProjection(LayerAnnotation):
    """
    Annotates a Batched fused linear projection layer, which applies
    the following transformation to inputs `X`:

    ``output = split(W @ X + b, output_dims, dim=-1)``
    """

    input_dim_field: str
    output_dims_field: str
    weight_param_name: str
    vec_dim_field: str


class AttentionInputTransformFused(FusedLinearProjection):
    """
    Annotates a layer which computes the query, key and value
    activations for an attention operation by projecting out
    the input activations.
    """


class BatchedAttentionInputTransformFused(BatchedFusedLinearProjection):
    """
    Annotates a layer which computes the query, key and value
    activations for an attention operation by projecting out
    the input activations using batched linear projections.
    """


class LinearProjection(LayerAnnotation):
    """
    Annotates a linear projection layer, which applies
    the following transformation to inputs `X`:

    ``output = W @ X + b``
    """

    input_dim_field: str = "in_features"
    output_dim_field: str = "out_features"
    weight_param_name: str = "weight"


class BatchedLinearProjection(LayerAnnotation):
    """
    Annotates a batched linear projection layer, which applies
    the following transformation to inputs `X`:

    ``output = W @ X + b``
    """

    input_dim_field: str = "in_features"
    output_dim_field: str = "out_features"
    vec_dim_field: str = "vec_dim"
    weight_param_name: str = "weight"


class AttentionInputTransformQuery(LinearProjection):
    """
    Annotates a layer which computes the query activation
    for attention operation.
    """


class AttentionInputTransformKey(LinearProjection):
    """
    Annotates a layer which computes the key activation
    for attention operation.
    """


class AttentionInputTransformValue(LinearProjection):
    """
    Annotates a layer which computes the value activation
    for attention operation.
    """


class AttentionOutputTransform(LinearProjection):
    """
    Annotates a layer which computes the output activation
    by projecting out the output of an attention operation
    in a transformer layer.
    """


class FeedForwardHiddenTransform(LinearProjection):
    """
    Annotates the feed-forward hidden transform projection
    layer of a transformer block.
    """


class FeedForwardOutputTransform(LinearProjection):
    """
    Annotates the feed-forward output transform projection
    layer of a transformer block.
    """


class BatchedAttentionInputTransformQuery(BatchedLinearProjection):
    """
    Annotates a layer which computes the query activation
    for attention operation using batched linear projection.
    """


class BatchedAttentionInputTransformKey(BatchedLinearProjection):
    """
    Annotates a layer which computes the key activation
    for attention operation using batched linear projection.
    """


class BatchedAttentionInputTransformValue(BatchedLinearProjection):
    """
    Annotates a layer which computes the value activation
    for attention operation using batched linear projection.
    """


class BatchedAttentionOutputTransform(BatchedLinearProjection):
    """
    Annotates a layer which computes the output activation
    by projecting out the output of an attention operation
    in a transformer layer using batched linear projection.
    """


class BatchedFeedForwardHiddenTransform(BatchedLinearProjection):
    """
    Annotates the feed-forward hidden transform Batched linear projection
    layer of a transformer block.
    """


class BatchedFeedForwardOutputTransform(BatchedLinearProjection):
    """
    Annotates the feed-forward output transform Batched linear projection
    layer of a transformer block.
    """


class MoEFeedForwardHiddenTransform(BatchedLinearProjection):
    """
    Annotates the feed-forward hidden transform projection
    layer of a mixture-of-experts transformer block.
    """


class MoEFeedForwardOutputTransform(BatchedLinearProjection):
    """
    Annotates the feed-forward output transform projection
    layer of a mixture-of-experts transformer block.
    """


def annotate_layer(module: _nn.Module, layer_annotations: _List[LayerAnnotationType]):
    """
    Mark a module with a list of :py:class:`LayerAnnotation` to indicate it is
    an adaptable layer of those types.
    """
    layer_types = getattr(module, _LAYER_ANNOTATION_FIELD, [])
    existing_layer_types = set(
        layer_type.__class__.__name__ for layer_type in layer_types
    )
    for layer_annotation in layer_annotations:
        if not 1 <= len(layer_annotation) <= 2:
            raise ValueError(
                f"Found invalid layer_annotation: {layer_annotation}. "
                f"It should be a tuple of 1 <= length <= 2."
            )

        if len(layer_annotation) == 2:
            layer_type_cls_name_or_cls, kwargs = layer_annotation
        else:
            # len(layer_annotation) == 1
            layer_type_cls_name_or_cls, kwargs = layer_annotation[0], None
        # we always use a string for getting the correct layer_type_cls
        # so that if a class is not in registry, we can catch it
        layer_type_cls_name = (
            layer_type_cls_name_or_cls.__name__
            if not isinstance(layer_type_cls_name_or_cls, str)
            else layer_type_cls_name_or_cls
        )
        if layer_type_cls_name in existing_layer_types:
            continue
        layer_type_cls: LayerAnnotation = (
            _LAYER_ANNOTATION_TYPE_REGISTRY.get_factory_fn(layer_type_cls_name)
        )
        layer_type = (
            layer_type_cls(**kwargs)  # pylint: disable=not-callable
            if kwargs
            else layer_type_cls()  # pylint: disable=not-callable
        )
        layer_types.append(layer_type)
    setattr(module, _LAYER_ANNOTATION_FIELD, layer_types)


def unannotate_layer(module: _nn.Module):
    """
    Removes layer annotations from a module.
    """
    if is_adaptable_layer(module):
        delattr(module, _LAYER_ANNOTATION_FIELD)


def is_adaptable_layer(module: _nn.Module) -> bool:
    """
    Returns ``True`` if a module supports getting adapted with an adapter,
    i.e., it has been annotated with a list of :py:class:`LayerAnnotation`.
    """
    return hasattr(module, _LAYER_ANNOTATION_FIELD)


def get_layer_annotations(
    module: _nn.Module,
    filter_types: _Optional[_Tuple[_Type[LayerAnnotation]]] = None,
) -> _Optional[_List[LayerAnnotation]]:
    """
    Returns the list of :py:class:`LayerAnnotation` annotation of a
    module, if it exists. If ``filter_types`` is specified, only returns
    layer types which are of specified
    """
    if is_adaptable_layer(module):
        layer_types = getattr(module, _LAYER_ANNOTATION_FIELD)
        if filter_types:
            return [lt for lt in layer_types if isinstance(lt, filter_types)]
        return layer_types
    return None



import logging as _logging

from tamm.layers.common import _BaseConfig
from tamm.models.common.metadata import ModelMetadata as _ModelMetadata

_logger = _logging.getLogger(__name__)


class ModelConfig(
    _BaseConfig,
    json_namespace="models",
):
    """
    The base class for model configs.  A :class:`.ModelConfig` is a
    :class:`.DataclassedPartial` with the model class as the target callable.

    For usage examples, please see the model
    configuration :ref:`user guide <configure_models>`.  Model developers should use the
    :class:`.ModelMixin` to create :class:`.ModelConfig` subclasses for new models.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # pylint: disable=access-member-before-definition
        if hasattr(self, "metadata") and self.metadata is None:
            self.metadata = _ModelMetadata()

    def create_model(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured instance of the model.

        Args:
            override_args: Optional positional arguments to override args specified in
                the config.  These args replace the first ``len(override_args)``
                positional args (and *all* varargs if ``override_args`` contains
                varargs) in the model's :meth:`~.ModelMixin.create_builder` method.
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overriden by
                ``override_args``.

        Returns:
            The newly created model.
        """
        return self._create(*override_args, **override_kwargs)

    @classmethod
    def _create_subclass_from_model(cls, model_cls):
        """
        Creates a :class:`.ModelConfig` for a model class which implements
        create_builder(*args, **kwargs)
        """
        config_cls = super()._create_subclass_from_module_cls(model_cls)
        config_cls.__doc__ = (
            "A :class:`.ModelConfig` subclass for configuring "
            f":class:`.{model_cls.__name__}` models.  Use the alias "
            f":attr:`.{model_cls.__name__}.Config` to access this class.  Please check "
            f":meth:`.{model_cls.__name__}.create_builder` for more "
            f"details about the  signature."
        )
        return config_cls

    def _to_json_dict_impl(self):
        result = super()._to_json_dict_impl()
        if self.metadata.is_empty:
            result.pop("metadata", None)
        return result

    @classmethod
    def _from_json_dict_impl(cls, **obj):
        obj.pop("__model_config_cls__", None)
        return super()._from_json_dict_impl(**obj)



from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch
from torch import nn as _nn

from tamm.generation.token_generators.base import TokenGenerator as _TokenGenerator
from tamm.generation.token_generators.next_token_logic import (
    NextTokenLogic as _NextTokenLogic,
)
from tamm.generation.token_generators.utils import (
    KVCacheType,
    check_eos_endings,
    create_random_number_generator,
    model_forward_step,
    pad_after_termination,
    prepare_input_ids,
    setup_kv_cache,
)


class _VanillaTokenGenerator(_TokenGenerator):
    """
    Token generator that uses autoregressive strategy for generating tokens,
    i.e., it generates first token by passing inputs to the model, and then appends this token
    to the inputs to generate the next token and so on.

    Args:
        model (:obj:`nn.Module`): The model instance.
        next_token_logic (:obj:`_NextTokenLogic`): Logic specifying how model's output logits should be
         used to generate the next token.
        eos_id (:obj:`Union[int, torch.Tensor, List[torch.Tensor]]`): End of sequence token ID(s).
        pad_id (:obj:`Union[int, torch.Tensor]`): Padding token ID.
        cache_type (:obj:`KVCacheType`, optional): Type of KV cache to use. Defaults to KVCacheType.V0.
        cache_dtype (:obj:`torch.dtype`, optional): Data type for KV cache. Defaults to None.
        output_logits (:obj:`bool`): When set to ``True``, :py:meth:`generate` returns logits
          corresponding to the generated ``token_ids``.
        output_probabilities (:obj:`bool`): When set to ``True``, :py:meth:`generate` returns probability
          distributions corresponding to the generated ``token_ids``.
    """

    def __init__(
        self,
        model: _nn.Module,
        next_token_logic: _NextTokenLogic,
        eos_id: _Union[int, _torch.Tensor, _List[_torch.Tensor]],
        pad_id: _Union[int, _torch.Tensor],
        cache_type: _Optional[KVCacheType] = KVCacheType.V0,
        cache_dtype: _Optional[_torch.dtype] = None,
        output_logits: bool = False,
        output_probabilities: bool = False,
    ):
        super().__init__(
            model=model,
            eos_id=eos_id,
            pad_id=pad_id,
        )
        self.next_token_logic = next_token_logic
        self._cache_type = cache_type
        self._cache_dtype = cache_dtype
        self.output_logits = output_logits
        self.output_probabilities = output_probabilities

    # pylint: disable=duplicate-code, too-many-locals
    @_torch.inference_mode()
    def _generate(
        self,
        input_ids: _torch.Tensor,
        max_new_tokens: int,
        seed: _Optional[int] = None,
    ) -> _Tuple[_torch.Tensor, _Optional[_torch.Tensor], _Optional[_torch.Tensor]]:
        input_ids = prepare_input_ids(input_ids)
        batch_size, prompt_len = input_ids.shape
        device = input_ids.device

        torch_rng = create_random_number_generator(device, seed)

        kv_cache = setup_kv_cache(
            model=self.model,
            batch_size=batch_size,
            cache_len=prompt_len + max_new_tokens,
            input_ids=input_ids,
            pad_id=self.pad_id,
            cache_type=self._cache_type,
            cache_dtype=self._cache_dtype,
        )

        eos_token_ids = [x.to(input_ids.device) for x in self.eos_ids]
        next_tokens = None
        return_logits, logits_list = None, []
        return_probs, probs_list = None, []

        for token_idx in range(max_new_tokens):
            output = model_forward_step(
                model=self.model,
                input_ids=input_ids,
                next_tokens=next_tokens,
                pad_id=self.pad_id,
                kv_cache=kv_cache,
                prompt_len=prompt_len,
                token_idx=token_idx,
            )
            logits = output.logits[:, -1, : self.model.config.vocab_size]
            next_token_output = self.next_token_logic(
                logits, return_probabilities=self.output_probabilities, rng=torch_rng
            )
            next_tokens = next_token_output.token_ids
            input_ids = _torch.cat([input_ids, next_tokens], dim=-1)

            if self.output_logits:
                logits_list.append(logits.unsqueeze(1))
            if self.output_probabilities:
                probs_list.append(next_token_output.probabilities.unsqueeze(1))  # type: ignore[union-attr]

            should_terminate = (
                check_eos_endings(
                    generated_tokens=next_tokens, eos_token_ids=eos_token_ids
                )
                .all()
                .item()
            )

            if should_terminate:
                break

        token_ids = input_ids[:, prompt_len:]
        token_ids = pad_after_termination(token_ids, self.pad_id, eos_token_ids)

        if self.output_logits:
            return_logits = _torch.cat(logits_list, dim=1)
        if self.output_probabilities:
            return_probs = _torch.cat(probs_list, dim=1)

        return token_ids, return_logits, return_probs



""" Defines a registry for tokenizers and provides utility functions for
creating, registering, listing, and validating tokenizers."""

import logging as _logging
from typing import Union as _Union

from tamm import _warnings as _tamm_warnings
from tamm._plugin import utils as plugin_utils
from tamm.runtime_configuration import rc
from tamm.tokenizers.common import PublishedTokenizerConfig, Tokenizer, TokenizerConfig
from tamm.tokenizers.registry import TokenizerRegistry, URIHandlerTokenizerRegistry

_logger = _logging.getLogger(__name__)


TokenizerSpecType = _Union[str, TokenizerConfig, PublishedTokenizerConfig, Tokenizer]


def get_tokenizer_registry() -> TokenizerRegistry:
    entrypoints = plugin_utils.discover_plugin_extras(
        rc.PLUGINS_EXTRAS_ENTRYPOINT_NAMES.tokenizer_registry
    )
    tokenizer_registry = URIHandlerTokenizerRegistry()
    for entrypoint in entrypoints:
        extra_registry = plugin_utils.execute_plugin_object_reference(
            entrypoint.value,
            call_obj=True,
        )
        tokenizer_registry += extra_registry

    return tokenizer_registry


def create_tokenizer_config(tokenizer_name: str, **kwargs):
    """
    Returns a tokenizer config from the tokenizer registry.
    Either `tokenizer_id` or `tokenizer_config_path` should be specified.

    Args:
        tokenizer_name (:obj:`str`): Either a `tamm` tokenizer's id,
            or a custom tokenizer config path.
    Returns:
        tokenizer_config: A tokenizer config object.
    """
    return get_tokenizer_registry().create_config(tokenizer_name, **kwargs)


def create_tokenizer(
    tokenizer_spec: TokenizerSpecType,
    *args,
    **kwargs,
) -> "Tokenizer":
    """
    Creates and returns a tokenizer instance based on the provided tokenizer spec.

    Args:
        tokenizer_spec (:obj:`str | TokenizerConfig | PublishedTokenizerConfig | TokenizerBase`):
            Acceptable input formats are
                1. A ``str`` of tokenizer_id, which is used to map tp a published tokenizer config.
                2. A :py:class:`.TokenizerConfig` object.
                3. A :py:class:`.PublishedTokenizerConfig` object.
                4. An instance of class inherited from :py:class:`.TokenizerBase`.

    Raises
        ValueError: If the model ID is not a string, is None. Also, if the
            registry has no entry for the requested tokenizer.
        KeyError: If the registry has no entry for the provided
            model ID.
    """
    return get_tokenizer_registry().create(
        tokenizer_spec,
        *args,
        **kwargs,
    )


def create_tokenizer_from_model_id(
    model_id: str,
    **kwargs,
) -> "Tokenizer":
    """
    Creates and returns a tokenizer instance based on the provided model ID.

    Args:
        model_id (:obj:`str`): The `tamm` model's id.
        vocab_path (:obj:`str`, optional): the path to the vocab file.
        user_defined_tokens (:obj:`list`, optional): List of user defined tokens.
            Defaults to `None`.

    Raises
        ValueError: If the model ID is not a string, is None. Also, if the
            registry has no entry for the requested tokenizer.
        KeyError: If the registry has no entry for the provided
            model ID.
    """
    if model_id is None or not isinstance(model_id, str):
        raise ValueError(
            "To create a tokenizer, model_id needs to be a valid string and cannot be "
            f"None or empty. Currently provided model_id = {model_id}"
        )
    try:
        # pylint: disable-next=import-outside-toplevel
        from tamm.model_repo import create_model_config

        config = create_model_config(model_id)
    except KeyError as e:
        raise KeyError(
            f"Cannot determine tokenizer because model ID '{model_id}' "
            f"does not exist"
        ) from e

    return get_tokenizer_registry().create(
        config.metadata.tokenizer_spec,
        **kwargs,
    )


def create_tokenizer_from_tokenizer_id(
    tokenizer_id: str,
    **kwargs,
) -> "Tokenizer":
    """
    Creates and returns a tokenizer instance based on the provided tokenizer ID.

    Raises ValueError: If the tokenizer ID is not a string, or is None.
        Also, if the registry has no entry for the requested tokenizer.
    """
    _tamm_warnings.deprecation(
        "create_tokenizer_from_tokenizer_id() is deprecated.  "
        "Please use `tamm.create_tokenizer` instead."
    )
    if tokenizer_id is None or not isinstance(tokenizer_id, str):
        raise ValueError(
            "To create a tokenizer, tokenizer_id needs to be a valid string and cannot "
            f"be None or empty. Currently provided tokenizer_id = {tokenizer_id}"
        )

    return get_tokenizer_registry().create(
        tokenizer_id,
        **kwargs,
    )


def list_tokenizers(include_descriptions=False, filter_deprecated=False):
    """
    Lists all tokenizer ids from the tokenizer registry.

    Args:
        include_descriptions (:obj:`bool`, optional): Specifies if descriptions
            should be included. If set to `True`, the return value will be a `dict`
            with tokenizer_ids as keys and descriptions as values. Defaults to `False`.
        filter_deprecated (:obj:`bool`, optional): True: don't show deprecated tokenizers;
            False: show deprecated tokenizers
    Returns:
        A `list` or a `dict` containing available tokenizers.
    """
    tokenizer_ids = get_tokenizer_registry().list_(filter_deprecated=filter_deprecated)
    if not include_descriptions:
        return tokenizer_ids
    return {
        tokenizer_id: get_tokenizer_registry().describe(tokenizer_id)
        for tokenizer_id in tokenizer_ids
    }



"""
adapters.model_adapters
=======================

.. autoclass:: tamm.adapters.ModelAdapter
    :show-inheritance:
    :members:
"""

# pylint: disable=no-member, cyclic-import

import dataclasses as _dataclasses
import logging as _logging
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm import _helpers
from tamm._adapters_v1.adapter_api import init as _init
from tamm._adapters_v1.adapter_api import (
    is_adapter_initialized as _is_adapter_initialized,
)
from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.utils import create_v1_state_dict, is_v0_lora_state_dict
from tamm.context_vars import resolve_device
from tamm.context_vars import resolve_pretrained_flag as _resolve_pretrained_flag
from tamm.utils import OptionalBool as _OptionalBool
from tamm.utils._pretrained import fetch_checkpoint as _fetch_checkpoint
from tamm.utils.json import JSONSerializableABCMixin as _JSONSerializableABCMixin

_logger = _logging.getLogger(__name__)


@_dataclasses.dataclass
class AdapterSpec:
    layer_adapters: _Optional[_Dict[str, _LayerAdapter]]


class ModelAdapter(_JSONSerializableABCMixin, json_namespace="adapters"):
    """
    Base class for model adapters.
    """

    # pylint: disable=signature-differs
    def __init_subclass__(cls):
        cls._inject_common_fields()
        _dataclasses.dataclass(cls)

    @classmethod
    def _inject_common_fields(cls):
        _helpers.set_annotated_class_attr(
            cls, "pretrained", _OptionalBool.NOTSET, _OptionalBool
        )
        _helpers.set_annotated_class_attr(cls, "pretrained_path", None, _Optional[str])
        _helpers.set_annotated_class_attr(cls, "device", None, _Optional[_torch.device])
        _helpers.set_annotated_class_attr(
            cls, "dtype", _torch.float32, _Optional[_torch.dtype]
        )
        _helpers.set_annotated_class_attr(cls, "freeze_params", False, bool)

    def create_adapters(self, model: _torch.nn.Module) -> AdapterSpec:
        """
        Create :py:class:`_LayerAdapter` for layers in model which are adapted
        by this model adapter.

        Args:
            model (:py:class:`torch.nn.Module`): Model to be adapted.

        Returns:
            An :py:class:`AdapterSpec` object which containts a dictionary
            mapping layer names to adapters used for adapting them.
        """
        with resolve_device(self.device):
            adapter_spec = self._create_adapters_impl(model)
        self._maybe_load_pretrained(adapter_spec)
        self._maybe_freeze_adapters(adapter_spec)
        return adapter_spec

    def adapt_model(self, model: _torch.nn.Module, adapter_id: str):
        """
        Adapt model with this model adapter using the specified ``adapter_id``.

        Args:
            model (:py:class:`_nn.Module`): Top level model to which adapters are to be added.
            adapter_id (:obj:`str`): A unique identifier for the adapters being added.
        """
        if not _is_adapter_initialized(model):
            model = _init(model)
        adapter_spec = self.create_adapters(model)
        for layer_name, layer_adapter in adapter_spec.layer_adapters.items():
            submodule = model.get_submodule(layer_name)
            submodule.add_adapter(adapter_id, layer_adapter)

    def _create_adapters_impl(self, model: _torch.nn.Module) -> AdapterSpec:
        """
        Implementation of create_adapters method.

        Create :py:class:`_LayerAdapter` for layers in model which are adapted
        by this model adapter.

        Args:
            model (:py:class:`torch.nn.Module`): Model to be adapted.

        Returns:
            An :py:class:`AdapterSpec` object which containts a dictionary
            mapping layer names to adapters used for adapting them.
        """
        raise NotImplementedError(
            "This method is not implemented by the base ModelAdapter class. "
        )

    @staticmethod
    def _transform_key(key: str, layer_name: str) -> str:
        """
        Remove adapter_id from layer_name to enable state dict loading.
        """
        layer_prefix = f"{layer_name}.adapters."
        return key[len(layer_prefix) :].split(".", 1)[1]

    def _maybe_load_pretrained(self, adapter_spec: AdapterSpec) -> None:
        """
        Load pretrained adapter weights if possible.
        """
        if self.pretrained_path is None:
            return
        resolved_pretrained_flag = _resolve_pretrained_flag(self.pretrained)
        if resolved_pretrained_flag != _OptionalBool.TRUE:
            return

        state_dict = _fetch_checkpoint(self.pretrained_path)
        if is_v0_lora_state_dict(state_dict):
            state_dict = create_v1_state_dict(state_dict, "V1_ADAPTER_ID_PLACE_HOLDER")

        for name, adapter in adapter_spec.layer_adapters.items():
            adapter_sd = {
                self._transform_key(k, layer_name=name): v
                for k, v in state_dict.items()
                if k.startswith(name)
            }
            adapter.load_state_dict(adapter_sd)

    def _maybe_freeze_adapters(self, adapter_spec: AdapterSpec) -> None:
        """
        Make adapters un-trainable if requested.
        """
        if not self.freeze_params:
            return
        for adapter in adapter_spec.layer_adapters.values():
            adapter.freeze()

    def _to_json_dict_impl(self):
        result = _helpers.dataclass_to_dict(
            self, omit_defaults=True  # only non-defaults for forward compatibility
        )
        return result

    @classmethod
    def _from_json_dict_impl(cls, **raw_dict: dict):
        raw_dict = _helpers.maybe_get_enum_member_for_all_str_enums(raw_dict, cls)
        return super()._from_json_dict_impl(**raw_dict)
